
import os
from pathlib import Path
from mako.template import Template
from pytos.securechange.xml_objects.restapi.step.access_request.designer import *
from pytos.securechange.xml_objects.restapi.step.access_request.verifier import *
from pytos.securechange.xml_objects.restapi.step.access_request.risk import *

logger = logging.getLogger(XML_LOGGER_NAME)
dir_path = list(Path(os.path.abspath(__file__)).parts[:-5])
dir_path.append('templates/risk_analysis_template.html')
RISK_ANALYSIS_HTML_PATH = os.path.join(*tuple(dir_path))


class Label(Flat_XML_Object_Base):
    def __init__(self, label):
        super().__init__(Elements.LABEL, content=label)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        label = xml_node.text
        return cls(label)


class Application_Target(XML_Object_Base):
    def __init__(self, application):
        self.application = [application]
        super().__init__(Elements.APPLICATION)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        application = get_xml_text_value(xml_node, Elements.APPLICATION)
        return cls(application)

    def to_pretty_str(self):
        if self.application[0] is not None:
            return "\n\t\tApplication: {}".format(self.application[0])
        else:
            return ""


class Named_Access_Request_Device(Target_Base):
    def __init__(self, num_id, object_name, object_type, object_details, management_name, management_id):

        self.object_name = object_name
        self.object_type = object_type
        self.object_details = object_details
        self.management_name = management_name
        self.management_id = management_id
        super().__init__(Elements.TARGET, num_id, TYPE_OBJECT)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, Elements.ID)
        object_name = get_xml_text_value(xml_node, Elements.OBJECT_NAME)
        object_type = get_xml_text_value(xml_node, Elements.OBJECT_TYPE)
        object_details = get_xml_text_value(xml_node, Elements.OBJECT_DETAILS)
        management_name = get_xml_text_value(xml_node, Elements.MANAGEMENT_NAME)
        management_id = get_xml_int_value(xml_node, Elements.MANAGEMENT_ID)
        return cls(num_id, object_name, object_type, object_details, management_name, management_id)

    def to_pretty_str(self):
        return "\n\t\tManagement Name: {}\n\t\tObject Name: {}\n\t\tObject Details: {}".format(self.management_name,
                                                                                               self.object_name,
                                                                                               self.object_details)

    def __str__(self):
        if all([self.management_name, self.object_name, self.object_details]):
            return "{}/{}/{}".format(self.management_name, self.object_name, self.object_details)
        else:
            return ""


class Any_Access_Request_Device(Target_Base):
    def __init__(self, num_id=None):
        self.management_name = "Any"
        super().__init__(Elements.TARGET, num_id, TYPE_ANY)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, Elements.ID)
        return cls(num_id)

    @staticmethod
    def to_pretty_str():
        return "\n\t\tManagement Name: Any"

    def __str__(self):
        return "Any"


class User_Target(XML_Object_Base):
    def __init__(self, user):
        self.user = [user]
        super().__init__(Elements.USER)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        user = get_xml_text_value(xml_node, Elements.USER)
        return cls(user)

    def to_pretty_str(self):
        if self.user[0] is not None:
            return "\n\t\tUser: {}".format(self.user[0])
        else:
            return ""

    def __str__(self):
        return ",".join(self.user)


class Network_Target(Access_Request_Target):
    def __init__(self, xml_tag, target_id, target_type, region):
        self.region = region
        super().__init__(xml_tag, target_id, target_type)

    def as_netaddr_obj(self):
        raise NotImplementedError

    def as_netaddr_set(self):
        """This returns a netaddr set representing the Network_Target"""
        return netaddr.IPSet(self.as_netaddr_obj())

    def to_pretty_str(self):
        raise NotImplementedError

    def from_xml_node(self, xml_node):
        raise NotImplementedError


class IP_Range_Access_Request_Target(Network_Target):
    def __init__(self, xml_tag, target_id, range_first_ip, range_last_ip, region):
        self.range_first_ip = range_first_ip
        self.range_last_ip = range_last_ip
        super().__init__(xml_tag, target_id, TYPE_RANGE, region)

    def to_pretty_str(self):
        target_string = "\n\t\tFirst IP Address: {}\n\t\tLast IP Address: {}".format(self.range_first_ip,
                                                                                     self.range_last_ip)
        return target_string

    def __str__(self):
        return "{}-{}".format(self.range_first_ip, self.range_last_ip)

    def as_netaddr_obj(self):
        """This returns a netaddr object representing the Ranged_Network_Target"""
        return netaddr.IPRange(self.range_first_ip, self.range_last_ip)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        range_first_ip = get_xml_text_value(xml_node, Elements.RANGE_FIRST_IP)
        range_last_ip = get_xml_text_value(xml_node, Elements.RANGE_LAST_IP)
        target_id = get_xml_int_value(xml_node, Elements.ID)
        region = get_xml_text_value(xml_node, Elements.REGION)
        return cls(xml_node.tag, target_id, range_first_ip, range_last_ip, region)


class IP_Access_Request_Target(Network_Target):
    def __init__(self, xml_tag, target_id, address, netmask, region, cidr=None):
        self.ip_address = address
        self.region = region

        # BUG: The netmask attribute is not always set since when creating an IPv6 target using the REST API
        # the netmask element must not exist (Can't even be empty)
        if netmask:
            self.netmask = netmask
        elif cidr:
            self.cidr = cidr

        super().__init__(xml_tag, target_id, TYPE_IP, region)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        target_id = get_xml_int_value(xml_node, Elements.ID)
        region = get_xml_text_value(xml_node, Elements.REGION)
        target_address = get_xml_text_value(xml_node, Elements.IP_ADDRESS)
        target_mask = get_xml_text_value(xml_node, Elements.NETMASK)
        target_cidr = get_xml_text_value(xml_node, Elements.CIDR)
        return cls(xml_node.tag, target_id, target_address, target_mask, region, target_cidr)

    def to_pretty_str(self):
        if hasattr(self, 'netmask'):
            return "\n\t\tIP Address: {}\n\t\tSubnet Mask: {}".format(self.ip_address, self.netmask)
        elif hasattr(self, 'cidr'):
            return "\n\t\tIP Address: {}\n\t\tCIDR: {}".format(self.ip_address, self.cidr)
        else:
            return "\n\t\tIP Address: {}".format(self.ip_address)

    def __str__(self):
        if hasattr(self, 'netmask'):
            return "{}/{}".format(self.ip_address, self.netmask)
        elif hasattr(self, 'cidr'):
            return "{}/{}".format(self.ip_address, self.cidr)
        else:
            return "{}".format(self.ip_address)

    def as_netaddr_obj(self):
        """This returns a netaddr object representing the Network_Target"""
        return netaddr.IPNetwork(self.__str__())


class IpAddress(XML_Object_Base):
    def __init__(self, ip_address):
        self.ip_address = ip_address
        super().__init__(Elements.IP_ADDRESS)

    @classmethod
    def from_xml_node(cls, xml_node):
        return cls(xml_node.text)

    def __str__(self):
        return self.ip_address


class DNS_Access_Request_Target(Network_Target):
    def __init__(self, xml_tag, target_id, address, host_name, region, dns_ip_addresses=None):
        # TODO: resolve dns if no ip address and only hostname
        self.region = region
        self.host_name = host_name
        if address:
            self.ip_address = address
        if dns_ip_addresses:
            self.dns_ip_addresses = dns_ip_addresses
        super().__init__(xml_tag, target_id, TYPE_DNS, region)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        target_id = get_xml_int_value(xml_node, Elements.ID)
        region = get_xml_text_value(xml_node, Elements.REGION)
        target_hostname = get_xml_text_value(xml_node, Elements.HOST_NAME)
        target_address = get_xml_text_value(xml_node, Elements.IP_ADDRESS)
        dns_ip_addresses = XML_List.from_xml_node_by_tags(xml_node, Elements.DNS_IP_ADDRESSES, Elements.IP_ADDRESS,
                                                          IpAddress, optional=True)
        return cls(xml_node.tag, target_id, target_address, target_hostname, region, dns_ip_addresses)

    def to_pretty_str(self):
        target_string = ""
        if hasattr(self, 'ip_address') and self.ip_address:
            target_string += "\n\t\tIP Address: {}".format(self.ip_address)
        elif hasattr(self, 'dns_ip_addresses'):
            target_string += "\n\t\tIP Addresses: {}".format(', '.join(ip.ip_address for ip in self.dns_ip_addresses))
        if self.host_name:
            target_string += "\n\t\tHostname: {}".format(self.host_name)
        return target_string

    def __str__(self):
        if hasattr(self, 'ip_address') and self.ip_address:
            return "{}/{}".format(self.host_name, self.ip_address)
        elif hasattr(self, 'dns_ip_addresses'):
            return "{}/{}".format(self.host_name, ', '.join(ip.ip_address for ip in self.dns_ip_addresses))

    def as_netaddr_obj(self):
        """This returns a netaddr object representing the Network_Target"""
        if hasattr(self, 'ip_address'):
            if not self.ip_address and self.host_name:
                return netaddr.IPNetwork(socket.gethostbyname(self.host_name))
            else:
                return netaddr.IPNetwork(self.ip_address)
        else:
            logger.info('Object {} has multiple DNS IP addresses, using the first one for conversion to netaddr obj'.format(self))
            if self.host_name:
                return netaddr.IPNetwork(socket.gethostbyname(self.host_name))
            else:
                return netaddr.IPNetwork(self.dns_ip_addresses.ip_address)


class Any_Access_Request_Target(Access_Request_Target):
    def __init__(self, xml_tag, num_id, region):
        super().__init__(xml_tag, num_id, TYPE_ANY, region)

    def __str__(self):
        return "Any"

    def to_pretty_str(self):
        return "\n\t\tIP Address: Any"

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        region = get_xml_text_value(xml_node, Elements.REGION)
        num_id = get_xml_int_value(xml_node, Elements.ID)
        return cls(xml_node.tag, num_id, region)

    @staticmethod
    def as_netaddr_obj():
        """This returns a netaddr object representing the Ranged_Network_Target"""
        return netaddr.IPNetwork("0.0.0.0/0")

    def as_netaddr_set(self):
        """This returns a netaddr set representing the Ranged_Network_Target"""
        return netaddr.IPSet(self.as_netaddr_obj())


class Object_Access_Request_Target(Access_Request_Target):
    def __init__(self, xml_tag, target_id, object_name, object_type, object_details, management_name, management_id,
                 object_UID, region):
        self.object_name = object_name
        self.object_type = object_type
        self.object_details = object_details
        self.object_UID = object_UID
        self.management_name = management_name
        self.management_id = management_id
        super().__init__(xml_tag, target_id, TYPE_OBJECT, region)

    def __str__(self):
        return "{}/{}".format(self.object_name, self.object_details)

    def to_pretty_str(self):
        object_string = ""
        if self.management_name:
            object_string += "\n\t\tManagement Name: {}".format(self.management_name)
        if self.object_name:
            object_string += "\n\t\tObject Name: {}".format(self.object_name)
        if self.object_details:
            object_string += "\n\t\tObject Details: {}".format(self.object_details)
        if self.object_UID:
            object_string += "\n\t\tObject UID: {}".format(self.object_UID)
        if self.object_type:
            object_string += "\n\t\tObject Type: {}".format(self.object_type)
        return object_string

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        target_id = get_xml_int_value(xml_node, Elements.ID)
        region = get_xml_text_value(xml_node, Elements.REGION)
        object_name = get_xml_text_value(xml_node, Elements.OBJECT_NAME)
        object_UID = get_xml_text_value(xml_node, Elements.OBJECT_UID)
        object_type = get_xml_text_value(xml_node, Elements.OBJECT_TYPE)
        object_details = get_xml_text_value(xml_node, Elements.OBJECT_DETAILS)
        management_name = get_xml_text_value(xml_node, Elements.MANAGEMENT_NAME)
        management_id = get_xml_int_value(xml_node, Elements.MANAGEMENT_ID)
        return cls(xml_node.tag, target_id, object_name, object_type, object_details, management_name, management_id,
                   object_UID, region)


class LDAP_Entity_Access_Request_Target(Access_Request_Target):
    def __init__(self, xml_tag, target_id, ldap_id, ldap_dn, ldap_name, region):
        self.ldap_entity_id = ldap_id
        self.ldap_entity_dn = ldap_dn
        self.ldap_entity_name = ldap_name
        super().__init__(xml_tag, target_id, TYPE_LDAP_ENTITY, region)

    def __str__(self):
        return "{}/{}".format(self.ldap_entity_name, self.ldap_entity_dn)

    def to_pretty_str(self):
        object_string = ""
        if self.ldap_entity_name:
            object_string += "\n\t\tLDAP Entity Name: {}".format(self.ldap_entity_name)
        if self.ldap_entity_dn:
            object_string += "\n\t\tLDAP Entity DN: {}".format(self.ldap_entity_dn)
        return object_string

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        target_id = get_xml_int_value(xml_node, Elements.ID)
        region = get_xml_text_value(xml_node, Elements.REGION)
        ldap_id = get_xml_text_value(xml_node, Elements.LDAP_ENTITY_ID)
        ldap_dn = get_xml_text_value(xml_node, Elements.LDAP_ENTITY_DN)
        ldap_name = get_xml_text_value(xml_node, Elements.LDAP_ENTITY_NAME)

        return cls(xml_node.tag, target_id, ldap_id, ldap_dn, ldap_name, region)


class Internet_Access_Request_Target(Access_Request_Target):
    def __init__(self, xml_tag, num_id, region):
        super().__init__(xml_tag, num_id, TYPE_INTERNET, region)

    def __str__(self):
        return "Internet"

    def to_pretty_str(self):
        return "\n\t\tIP Address: Internet"

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        region = get_xml_text_value(xml_node, Elements.REGION)
        num_id = get_xml_int_value(xml_node, Elements.ID)
        return cls(xml_node.tag, num_id, region)


class Service_Target(Access_Request_Target):
    def __init__(self, service_id=None, service_type=None):
        super().__init__(Elements.SERVICE, service_id, service_type)

    def to_pretty_str(self):
        raise NotImplementedError

    def from_xml_node(self, xml_node):
        raise NotImplementedError


class Any_Service_Target(Service_Target):
    def __init__(self, num_id=None):
        super().__init__(num_id, TYPE_ANY)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, Elements.ID)
        return cls(num_id)

    def to_pretty_str(self):
        return "\n\t\tService: Any"

    def __str__(self):
        return "Any"


class Protocol_Service_Target(Service_Target):
    def __init__(self, num_id, port, protocol, service_type):
        self.port = port
        self.protocol = protocol
        self.type = service_type
        super().__init__(num_id, SERVICE_OBJECT_TYPE_PROTOCOL)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, Elements.ID)
        protocol = get_xml_text_value(xml_node, Elements.PROTOCOL)
        port = get_xml_text_value(xml_node, Elements.PORT)
        service_type = get_xml_text_value(xml_node, Elements.TYPE)
        return cls(num_id, port, protocol, service_type)

    def to_pretty_str(self):
        return "\n\t\tProtocol: {}\n\t\tPort: {}".format(self.protocol, self.port)

    def __str__(self):
        return "{} {}".format(self.protocol, self.port)


class ApplicationPredefinedServiceTarget(Service_Target):
    def __init__(self, service_id, application_name, services):
        self.id = service_id
        self.application_name = application_name
        self.services = services
        super().__init__(service_id, SERVICE_OBJECT_TYPE_APPLICATION_IDENTITY)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        service_id = get_xml_int_value(xml_node, Elements.ID)
        application_name = get_xml_text_value(xml_node, Elements.APPLICATION_NAME)
        service_type_class_dict = {SERVICE_OBJECT_TYPE_PREDEFINED: Predefined_Service_Target,
                                   SERVICE_OBJECT_TYPE_PROTOCOL: Protocol_Service_Target,
                                   TYPE_OBJECT: Object_Access_Request_Target, TYPE_ANY: Any_Service_Target}
        services = XML_List.from_xml_node_by_type_dict(xml_node, Elements.SERVICES, Elements.SERVICE,
                                                       service_type_class_dict, True)
        return cls(service_id, application_name, services)

    def to_pretty_str(self):
        return "\n\t\tPredefined Appliciation Name: {}".format(self.application_name)

    def __str__(self):
        return "Predefined Appliciation {}".format(self.application_name)


class Predefined_Service_Target(Service_Target):
    def __init__(self, num_id, protocol, protocol_type, predefined_name, port=None):
        self.protocol = protocol
        self.predefined_name = predefined_name
        self.type = protocol_type
        if port:
            self.port = port
        super().__init__(num_id, SERVICE_OBJECT_TYPE_PREDEFINED)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, Elements.ID)
        protocol = get_xml_text_value(xml_node, Elements.PROTOCOL)
        predefined_name = get_xml_text_value(xml_node, Elements.PREDEFINED_NAME)
        protocol_type = get_xml_text_value(xml_node, Elements.TYPE)
        port = get_xml_text_value(xml_node, Elements.PORT)
        return cls(num_id, protocol, protocol_type, predefined_name, port)

    def __str__(self):
        return self.predefined_name

    def to_pretty_str(self):
        return "\n\t\tPredefined name: {}\n\t\t Protocol: {}".format(self.predefined_name, self.protocol)


class Access_Request(XML_Object_Base):
    def __init__(self, order, targets, users, sources, destinations, services, applications, action, comment, labels,
                 verifier_result, risk_analysis_result, ar_id, source_domain=None, destination_domain=None,
                 use_topology=None):
        self.id = ar_id
        self.order = order
        self.targets = XML_List(Elements.TARGETS, targets)
        self.users = XML_List(Elements.USERS, users)
        self.sources = XML_List(Elements.SOURCES, sources)
        self.destinations = XML_List(Elements.DESTINATIONS, destinations)
        self.services = XML_List(Elements.SERVICES, services)
        self.applications = XML_List(Elements.APPLICATIONS, applications)
        self.action = action
        self.comment = comment
        self.use_topology = use_topology
        self.labels = XML_List(Elements.LABELS, labels)
        self.verifier_result = verifier_result
        self.risk_analysis_result = risk_analysis_result
        if source_domain and destination_domain:
            self.source_domain = source_domain
            self.destination_domain = destination_domain
        super().__init__(Elements.ACCESS_REQUEST)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        order = get_xml_text_value(xml_node, Elements.ORDER)
        ar_id = get_xml_int_value(xml_node, Elements.ID)
        action = get_xml_text_value(xml_node, Elements.ACTION)
        comment = get_xml_text_value(xml_node, Elements.COMMENT)
        use_topology = get_xml_text_value(xml_node, Elements.USE_TOPOLOGY)
        verifier_result_node = get_xml_node(xml_node, Elements.VERIFIER_RESULT, True)
        if verifier_result_node is not None:
            verifier_result = Verifier_Result.from_xml_node(verifier_result_node)
        else:
            verifier_result = None

        risk_analysis_result_node = get_xml_node(xml_node, Elements.RISK_ANALYSIS_RESULT, True)
        if risk_analysis_result_node is not None:
            risk_analysis_result = Risk_Analysis_Result.from_xml_node(risk_analysis_result_node)
        else:
            risk_analysis_result = None

        labels = XML_List.from_xml_node_by_tags(xml_node, Elements.LABELS, Elements.LABEL, Label, True)
        applications = XML_List.from_xml_node_by_tags(xml_node, Elements.APPLICATIONS, Elements.APPLICATION,
                                                      Application_Target, True)
        device_type_to_class_dict = {TYPE_ANY: Any_Access_Request_Device, TYPE_OBJECT: Named_Access_Request_Device}
        targets = XML_List.from_xml_node_by_type_dict(xml_node, Elements.TARGETS, Elements.TARGET,
                                                      device_type_to_class_dict)
        users = XML_List.from_xml_node_by_tags(xml_node, Elements.USERS, Elements.USER, User_Target, True)

        target_type_to_class_dict = {TYPE_RANGE: IP_Range_Access_Request_Target, TYPE_IP: IP_Access_Request_Target,
                                     TYPE_DNS: DNS_Access_Request_Target, TYPE_OBJECT: Object_Access_Request_Target,
                                     TYPE_ANY: Any_Access_Request_Target, TYPE_INTERNET: Internet_Access_Request_Target,
                                     TYPE_LDAP_ENTITY: LDAP_Entity_Access_Request_Target}

        service_type_class_dict = {SERVICE_OBJECT_TYPE_PREDEFINED: Predefined_Service_Target,
                                   SERVICE_OBJECT_TYPE_PROTOCOL: Protocol_Service_Target,
                                   TYPE_OBJECT: Object_Access_Request_Target, TYPE_ANY: Any_Service_Target,
                                   SERVICE_OBJECT_TYPE_APPLICATION_IDENTITY: ApplicationPredefinedServiceTarget}

        sources = XML_List.from_xml_node_by_type_dict(xml_node, Elements.SOURCES, Elements.SOURCE,
                                                      target_type_to_class_dict, True)
        destinations = XML_List.from_xml_node_by_type_dict(xml_node, Elements.DESTINATIONS, Elements.DESTINATION,
                                                           target_type_to_class_dict, True)
        services = XML_List.from_xml_node_by_type_dict(xml_node, Elements.SERVICES, Elements.SERVICE,
                                                       service_type_class_dict, True)
        source_domain = get_xml_text_value(xml_node, Elements.SOURCE_DOMAIN)
        destination_domain = get_xml_text_value(xml_node, Elements.DESTINATION_DOMAIN)
        return cls(order, targets, users, sources, destinations, services, applications, action, comment, labels,
                   verifier_result, risk_analysis_result, ar_id, source_domain, destination_domain, use_topology)

    def to_pretty_str(self):
        access_request_string = "Access Request {}:\n".format(self.order)

        access_request_string += "\tTargets: "
        for target in self.targets:
            access_request_string += target.to_pretty_str()
        access_request_string += "\n\tSources: "
        for source in self.sources:
            access_request_string += source.to_pretty_str()
        access_request_string += "\n\tDestinations: "
        for destination in self.destinations:
            access_request_string += destination.to_pretty_str()
        access_request_string += "\n\tServices: "
        for service in self.services:
            access_request_string += service.to_pretty_str()
        if self.applications and any((application.to_pretty_str() for application in self.applications)):
            access_request_string += "\n\tApplications: "
            for application in self.applications:
                access_request_string += application.to_pretty_str()
        if self.users and any((user.to_pretty_str() for user in self.users)):
            access_request_string += "\n\tUsers: "
            for user in self.users:
                access_request_string += user.to_pretty_str()
        if self.comment is not None:
            access_request_string += "\n" + textwrap.fill("\tComment: {}".format(unescape(self.comment)),
                                                          initial_indent='', subsequent_indent='\t\t ')
        access_request_string += "\n"
        return access_request_string

    def __str__(self):
        return self.to_pretty_str()

    def to_csv_row(self, *, delimiter=",", inline_delimiter=";", export_targets=False):
        """ This will generate the CSV row in SC format.

        :param delimiter: The delimiter between fields (sources, destinations and etc)
        :type delimiter: str
        :param inline_delimiter: The delimiter in field itself for multiple values
        "type inline_delimiter: str
        :param export_targets: If to include targets (not compatible with SC import format)
        :type export_targets: bool
        :return: str
        """
        sources = inline_delimiter.join(str(src) for src in self.sources if src)
        destinations = inline_delimiter.join(str(dst) for dst in self.destinations if dst)
        services = inline_delimiter.join(str(srv) for srv in self.services if srv)
        action = self.action if self.action else ""
        comment = '"{}"'.format(self.comment.replace("\n", "\\n").replace('"', "'")) if self.comment else ""
        if export_targets:
            targets = inline_delimiter.join(str(target) for target in self.targets if target)
            row_data = (targets, sources, destinations, services, action, comment)
        else:
            row_data = (sources, destinations, services, action, comment)
        return delimiter.join(row_data)

    def get_risk_analysis_result_as_html(self, html_template_path=RISK_ANALYSIS_HTML_PATH, application_name=None,
                                         as_html=True):
        """Returns the risk analysis result as formatted html table or as a tuple

        :param html_template_path:
        :param application_name:
        :param as_html: If True a formatted html string is returned. if False, a tuple is returned
        :type as_html: bool
        :return:
        :rtype: str|tuple
        """
        def get_string_of_resources(resources):
            items = []
            for resource in resources:
                if isinstance(resource, (Violation_Any_Source, Violation_Any_Destination, Violation_Any_Service))\
                        or resource is None:
                    items.append("Any")
                elif isinstance(resource, (Violation_Not_Allowed_Group_Member_service_Object, Violation_Allowed_Group_Member_service_Object)):
                    items.append(resource.group_member_path)
                else:
                    items.append(resource.name)
            return ', '.join(items)

        rows = []
        for security_policy_violation in self.risk_analysis_result.security_policy_violations:
            severity = security_policy_violation.severity
            matrix = security_policy_violation.matrix_cell_violation
            if matrix:
                sources = get_string_of_resources(matrix.sources)
                violations = "Sources in zone {}: {}<BR>".format(matrix.from_zone, sources)
                destinations = get_string_of_resources(matrix.destinations)
                violations += "Destinations in zone {}: {}<BR>".format(matrix.to_zone, destinations)
                violations += "-----------------------------------------------<BR>"
                allowed_services = ""
                if matrix.allowed_services:
                    allowed_services = get_string_of_resources(matrix.allowed_services)
                not_allowed_services = "All services"
                if matrix.not_allowed_services:
                    not_allowed_services = get_string_of_resources(matrix.not_allowed_services)
                violations += "Violating services: {}".format(not_allowed_services)
                security_requirement = 'Policy control "{}" (Global Security Zone Matrix)<BR>'.format(
                    security_policy_violation.security_zone_matrix.name)
                security_requirement += '{} -> {} (Block all)<BR>'.format(matrix.from_zone, matrix.to_zone,
                                                                          matrix.destinations)
                security_requirement += "-----------------------------------------------<BR>"
                security_requirement += 'Services allowed: {}<BR>'.format(allowed_services)
                rows.append((severity, violations, security_requirement))

        compliance_policies = []
        for compliance_policy in self.risk_analysis_result.compliance_policies:
            rules = [(compliance_rule.number, compliance_rule.name) for compliance_rule in
                     compliance_policy.compliance_rules]
            compliance_policies.append((compliance_policy.name, compliance_policy.type, rules))

        if as_html:
            ar_header = "Severity Violations Security-Requirement".split()
            compliance_policy_header = "Name Type Rule".split()
            template = Template(filename=html_template_path)
            return template.render(headers=ar_header, ars=[(self.order, rows, compliance_policies)],
                                   compliance_headers=compliance_policy_header,
                                   app_name=application_name)

        return self.order, rows, compliance_policies


class Step_Field_Multi_Access_Request(Step_Multi_Field_Base):
    FIELD_CONTENT_ATTRIBUTES = "access_requests"

    def __init__(self, num_id, name, access_requests=None, designer_result=None, read_only=None):
        if access_requests is None:
            access_requests = []
        self.access_requests = access_requests
        if designer_result is not None:
            self.designer_result = designer_result
        super().__init__(num_id, name, read_only)
        self.set_attrib(Attributes.XSI_TYPE, Attributes.FIELD_TYPE_MULTI_ACCESS_REQUEST)

    def get_access_request_by_index(self, access_request_index):
        access_request_list = sorted(self.access_requests, key=lambda ar: ar.order)
        return access_request_list[access_request_index]

    def get_next_access_request_order(self):
        return "AR{}".format(len(self.access_requests) + 1)

    def get_all_sources(self):
        return itertools.chain.from_iterable((ar.sources for ar in self.access_requests))

    def get_all_destinations(self):
        return itertools.chain.from_iterable((ar.destinations for ar in self.access_requests))

    def get_all_services(self):
        return itertools.chain.from_iterable((ar.services for ar in self.access_requests))

    def get_all_targets(self):
        return itertools.chain.from_iterable((ar.targets for ar in self.access_requests))

    def get_all_applications(self):
        return itertools.chain.from_iterable((ar.applications for ar in self.access_requests))

    def get_all_users(self):
        return itertools.chain.from_iterable((ar.users for ar in self.access_requests))

    def get_all_verifier_results(self):
        return (ar.verifier_result for ar in self.access_requests)

    def get_all_risk_analysis_results(self):
        return (ar.risk_analysis_result for ar in self.access_requests)

    def get_all_risk_analysis_results_as_html(self, html_template_path=RISK_ANALYSIS_HTML_PATH,
                                              application_name=None, risk=None):
        """Returns all risks as formatted html

        :param html_template_path:
        :param application_name: The relevant SecureApp application name. Usually obtained by ticket.application_details.name
        :param risk: type of risk to export. Can be one of None, Risk_Analysis_Result.HAS_RISK or Risk_Analysis_Result.MANUALLY_DISREGARDED
        :return:
        """

        valid_risk_values = (Risk_Analysis_Result.HAS_RISK, Risk_Analysis_Result.MANUALLY_DISREGARDED, None)
        if risk not in valid_risk_values:
            raise ValueError("'risk' param must be one of {}".format(valid_risk_values))

        if risk is None:
            risk_funcs = (Risk_Analysis_Result.has_risk, Risk_Analysis_Result.is_manually_disregarded)
        elif risk == Risk_Analysis_Result.HAS_RISK:
            risk_funcs = (Risk_Analysis_Result.has_risk,)
        elif risk == Risk_Analysis_Result.MANUALLY_DISREGARDED:
            risk_funcs = (Risk_Analysis_Result.is_manually_disregarded,)

        ar_header = "Severity Violations Security-Requirement".split()
        compliance_policy_header = "Name Type Rule".split()
        ars_with_risk = [ar.get_risk_analysis_result_as_html(as_html=False) for ar in self.access_requests
                         if any(risk_func(ar.risk_analysis_result) for risk_func in risk_funcs)]
        template = Template(filename=html_template_path)
        return template.render(headers=ar_header, ars=ars_with_risk, compliance_headers=compliance_policy_header,
                               app_name=application_name)

    def get_designer_results(self, username, password):
        if not hasattr(self, 'designer_result'):
            return None
        designer_result_url = self.designer_result.get_result_url()
        url_parse = requests.utils.urlparse(designer_result_url)
        url_path = url_parse.path
        if "multi_access_request" not in url_path:  # fix bug in SC API
            url_path = "{}multi_access_request/designer".format(url_parse.path.split("fields")[0])
        try:
            url_helper = Secure_API_Helper(url_parse.netloc, (username, password))
            response_string = url_helper.get_uri(url_path, expected_status_codes=200).response.content
        except (REST_Service_Unavailable_Error, requests.RequestException) as e:
            message = "Failed to GET designer results"
            logger.error(message)
            raise IOError(message)
        return DesignerResults.from_xml_string(response_string)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        field_name = get_xml_text_value(xml_node, Elements.NAME)
        field_id = get_xml_int_value(xml_node, Elements.ID)
        field_read_only = get_xml_text_value(xml_node, Elements.READ_ONLY)
        access_requests = []
        for access_request_node in xml_node.findall(Elements.ACCESS_REQUEST):
            access_request = Access_Request.from_xml_node(access_request_node)
            access_requests.append(access_request)

        designer_result_node = get_xml_node(xml_node, Elements.DESIGNER_RESULT, True)
        if designer_result_node is not None:
            designer_result = DesignerResult.from_xml_node(designer_result_node)
        else:
            designer_result = None

        return cls(field_id, field_name, access_requests, designer_result, field_read_only)

    def to_pretty_str(self):
        output = "Access request field '{}'\n:".format(self.name)
        for ar in self.access_requests:
            output += "\n{}\n".format(ar.to_pretty_str())
        return output

    def to_csv(self, *, delimiter=",", inline_delimiter=";", export_targets=False):
        return "\n".join(
                ar.to_csv_row(delimiter=delimiter, inline_delimiter=inline_delimiter, export_targets=export_targets) for
                ar in self.access_requests)



import enum


@enum.unique
class ConnectionStatus(enum.Enum):
    NOT_APPLICABLE = "NOT_APPLICABLE"
    CALCULATING = "CALCULATING"
    CONNECTED = "CONNECTED"
    DISCONNECTED = "DISCONNECTED"
    DOMAIN_DELETED = "DOMAIN_DELETED"
    NOT_COMPLETE = "NOT_COMPLETE"
    USER_IN_SOURCE = "USER_IN_SOURCE"
    L7_APPLICATION = "L7_APPLICATION"
    ANY_OR_A_CLASS = "ANY_OR_A_CLASS"
    INTERNET = "INTERNET"
    NON_CONTINUOUS_MASK = "NON_CONTINUOUS_MASK"




import logging
import netaddr

from pytos.common.base_types import XML_Object_Base, XML_List, XSI_Object, Single_Service_Type, Range_Service_Type, \
    Group_Service_Type, Any_Service_Type, Comparable, IPNetworkMixin, Flat_XML_Object_Base
from pytos.common.definitions import xml_tags
from pytos.common.definitions.xml_tags import Attributes
from pytos.common.functions import str_to_bool, get_iana_protocols, netmask_to_cidr, XML_LOGGER_NAME
from pytos.common.functions.xml import get_xml_text_value, get_xml_int_value, get_xml_node, \
    create_tagless_xml_objects_list
from pytos.securetrack.xml_objects.base_types import Base_Object, Network_Object, Service
from pytos.securetrack.xml_objects.rest.device import Device, Device_Revision

logger = logging.getLogger(XML_LOGGER_NAME)


class Bindings_List(XML_List):
    def __init__(self, bindings):
        """
        :type bindings: list[Rule_Binding]
        """
        self.bindings = bindings
        super().__init__(xml_tags.Elements.BINDINGS, bindings)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        bindings = []
        for binding_node in xml_node.iter(tag=xml_tags.Elements.BINDING):
            bindings.append(Rule_Binding.from_xml_node(binding_node))
        return cls(bindings)


class Cleanup_Set(XML_Object_Base):
    def __init__(self, shadowed_rules_cleanup=None):
        self.shadowed_rules_cleanup = shadowed_rules_cleanup
        super().__init__(xml_tags.Elements.CLEANUP_SET)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        shadowed_rules_cleanup = Shadowed_Rules_Cleanup.from_xml_node(
                get_xml_node(xml_node, xml_tags.Elements.SHADOWED_RULES_CLEANUP))
        return cls(shadowed_rules_cleanup)


class Rules_List(XML_List):
    def __init__(self, count, total, rules):
        super().__init__(xml_tags.Elements.RULES, rules)
        self.count = count
        self.total = total

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        count = get_xml_int_value(xml_node, xml_tags.Elements.COUNT)
        total = get_xml_int_value(xml_node, xml_tags.Elements.TOTAL)
        rules = []
        for rule_node in xml_node.iter(tag=xml_tags.Elements.RULE):
            rules.append(Rule.from_xml_node(rule_node))
        return cls(count, total, rules)


class Policy_List(XML_List):
    def __init__(self, policies):
        """
        :type policies: list[Policy]
        """
        super().__init__(xml_tags.Elements.POLICIES, policies)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        policies = []
        for policy_node in xml_node.iter(tag=xml_tags.Elements.POLICY):
            policies.append(Policy.from_xml_node(policy_node))
        return cls(policies)


class Shadowed_Rules_Cleanup(XML_Object_Base):
    def __init__(self, shadowed_rules=None):
        self.shadowed_rules = XML_List(xml_tags.Elements.SHADOWED_RULES, shadowed_rules)
        super().__init__(xml_tags.Elements.SHADOWED_RULES_CLEANUP)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        shadowed_rules = XML_List(xml_tags.Elements.SHADOWED_RULES)
        for shadowed_rule_node in xml_node.iter(tag=xml_tags.Elements.SHADOWED_RULE):
            shadowed_rules.append(Shadowed_Rule.from_xml_node(shadowed_rule_node))
        return cls(shadowed_rules)


class Shadowed_Rule(XML_Object_Base):
    def __init__(self, rule, shadowing_rules=None):
        self.rule = rule
        self.shadowing_rules = shadowing_rules
        super().__init__(xml_tags.Elements.SHADOWED_RULE)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        rule = Rule.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.RULE))
        shadowing_rules = []

        shadowing_rules_node = get_xml_node(xml_node, xml_tags.Elements.SHADOWING_RULES, True)
        if shadowing_rules_node:
            for rule_node in shadowing_rules_node.iter(tag=xml_tags.Elements.RULE):
                shadowing_rules.append(Rule.from_xml_node(rule_node))
        else:
            shadowing_rules = None
        return cls(rule, shadowing_rules)


class Record_Set(XML_Object_Base):
    TIME_DATE_FORMAT_STRING = "%Y-%m-%dT%H:%M:%SZ"

    def __init__(self, businessowneremail, businessownername, expiredate, record_id, ticketcr, automatic=None):
        self.businessOwnerEmail = businessowneremail
        self.businessOwnerName = businessownername
        self.expireDate = expiredate
        self.id = record_id
        self.ticketCr = ticketcr
        self.automatic = automatic
        super().__init__(xml_tags.Elements.RECORD_SET)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        businessowneremail = get_xml_text_value(xml_node, xml_tags.Elements.BUSINESSOWNEREMAIL)
        businessownername = get_xml_text_value(xml_node, xml_tags.Elements.BUSINESSOWNERNAME)
        expiredate = get_xml_text_value(xml_node, xml_tags.Elements.EXPIREDATE)
        record_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        ticketcr = get_xml_text_value(xml_node, xml_tags.Elements.TICKETCR)
        automatic = get_xml_text_value(xml_node, xml_tags.Elements.AUTOMATIC)
        return cls(businessowneremail, businessownername, expiredate, record_id, ticketcr, automatic)

    def set_date_as_expiry_date(self, date):
        new_expiry_date = date.strftime(Record_Set.TIME_DATE_FORMAT_STRING)
        self.expireDate = new_expiry_date


class Rule_Documentation(XML_Object_Base):
    def __init__(self, tech_owner, comment, record_sets, secure_app_applications, new_rule=False):
        self.tech_owner = tech_owner
        self.comment = comment
        self.record_sets = record_sets
        self.secure_app_applications = secure_app_applications
        if new_rule:
            super().__init__(xml_tags.Elements.RULE_DOCUMENTATION)
        else:
            super().__init__(xml_tags.Elements.DOCUMENTATION)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        tech_owner = get_xml_text_value(xml_node, xml_tags.Elements.TECH_OWNER)
        comment = get_xml_text_value(xml_node, xml_tags.Elements.COMMENT)
        record_sets = []
        for record_set_node in xml_node.iter(tag=xml_tags.Elements.RECORD_SET):
            record_sets.append(Record_Set.from_xml_node(record_set_node))
        secure_app_applications = []
        for secure_app_application_node in xml_node.iter(tag=xml_tags.Elements.SECURE_APP_APPLICATION):
            secure_app_applications.append(SecureApp_Application.from_xml_node(secure_app_application_node))

        return cls(tech_owner, comment, record_sets, secure_app_applications)


class Application(XML_Object_Base):
    def __init__(self, app_id, display_name, name):
        self.id = app_id
        self.display_name = display_name
        self.name = name
        super().__init__(xml_tags.Elements.APPLICATION)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
         Initialize the object from a XML node.
         :param xml_node: The XML node from which all necessary parameters will be parsed.
         :type xml_node: xml.etree.Element
         """
        app_id = get_xml_text_value(xml_node, xml_tags.Elements.ID)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        return cls(app_id, display_name, name)

    def __str__(self):
        return self.display_name


class Rule(XML_Object_Base, Comparable):
    def __init__(self, num_id, uid, cp_uid, order, binding, action, comment, dst_networks, dst_networks_negated,
                 dst_services, dst_services_negated, disabled, external, name, rule_number, src_networks,
                 src_networks_negated, src_services_negated, track, rule_type, documentation, device_id, implicit,
                 application, vpn, rule_text=None):
        self.id = num_id
        self.uid = uid
        self.cp_uid = cp_uid
        self.order = order
        self.binding = binding
        self.action = action
        self.comment = comment
        self.dst_networks = dst_networks
        self.dst_networks_negated = dst_networks_negated
        self.dst_services = dst_services
        self.dst_services_negated = dst_services_negated
        self.disabled = disabled
        self.external = external
        self.name = name
        self.rule_number = rule_number
        self.src_networks = src_networks
        self.src_networks_negated = src_networks_negated
        self.src_services_negated = src_services_negated
        self.track = track
        self.type = rule_type
        self.documentation = documentation
        self.device_id = device_id
        self.implicit = implicit
        self._network_id_to_object = {}
        self.application = application
        self.vpn = vpn
        self.rule_text = rule_text
        super().__init__(xml_tags.Elements.RULE)

    def _key(self):
        hash_keys = [self.id, self.uid]
        if self.device_id:
            hash_keys.append(self.device_id)
        if self.binding:
            try:
                hash_keys.append(self.binding.uid)
            except AttributeError:
                pass
        return tuple(hash_keys)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        uid = get_xml_text_value(xml_node, xml_tags.Elements.UID)
        cp_uid = get_xml_text_value(xml_node, xml_tags.Elements.CP_UID)
        order = get_xml_text_value(xml_node, xml_tags.Elements.ORDER)
        binding_node = get_xml_node(xml_node, xml_tags.Elements.BINDING, True)
        if binding_node:
            binding = Rule_Binding.from_xml_node(binding_node)
        else:
            binding = None
        action = get_xml_text_value(xml_node, xml_tags.Elements.ACTION)
        comment = get_xml_text_value(xml_node, xml_tags.Elements.COMMENT)
        dst_networks_negated = get_xml_text_value(xml_node, xml_tags.Elements.DEST_NETWORKS_NEGATED)
        dst_services = create_tagless_xml_objects_list(xml_node, xml_tags.Elements.DST_SERVICE, Destination_Service)
        dst_services_negated = get_xml_text_value(xml_node, xml_tags.Elements.DEST_SERVICES_NEGATED)
        disabled = get_xml_text_value(xml_node, xml_tags.Elements.DISABLED)
        external = get_xml_text_value(xml_node, xml_tags.Elements.EXTERNAL)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        rule_number = get_xml_text_value(xml_node, xml_tags.Elements.RULE_NUMBER)
        src_networks = create_tagless_xml_objects_list(xml_node, xml_tags.Elements.SRC_NETWORK, Source_Network)
        dst_networks = create_tagless_xml_objects_list(xml_node, xml_tags.Elements.DST_NETWORK, Destination_Network)
        src_networks_negated = get_xml_text_value(xml_node, xml_tags.Elements.SRC_NETWORKS_NEGATED)
        src_services_negated = get_xml_text_value(xml_node, xml_tags.Elements.SRC_SERVICES_NEGATED)
        track_node = get_xml_node(xml_node, xml_tags.Elements.TRACK, True)
        if track_node is not None:
            track = Rule_Track.from_xml_node(track_node)
        else:
            track = None
        rule_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        documentation_node = get_xml_node(xml_node, xml_tags.Elements.DOCUMENTATION, True)
        if documentation_node:
            documentation = Rule_Documentation.from_xml_node(documentation_node)
        else:
            documentation = None
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        implicit = get_xml_text_value(xml_node, xml_tags.Elements.IMPLICIT)
        application_node = get_xml_node(xml_node, xml_tags.Elements.APPLICATION, True)
        vpn = create_tagless_xml_objects_list(xml_node, xml_tags.Elements.VPN, RuleVPNOption)
        rule_text = get_xml_text_value(xml_node, xml_tags.Elements.RULE_TEXT)
        if application_node:
            application = Application.from_xml_node(application_node)
        else:
            application = None

        return cls(num_id, uid, cp_uid, order, binding, action, comment, dst_networks, dst_networks_negated,
                   dst_services, dst_services_negated, disabled, external, name, rule_number, src_networks,
                   src_networks_negated, src_services_negated, track, rule_type, documentation, device_id, implicit,
                   application, vpn, rule_text)

    def __str__(self):
        src_negated, dst_negated, srv_negated = "", "", ""
        if self.src_networks_negated and str_to_bool(self.src_networks_negated):
            src_negated = "NOT "
        if self.dst_networks_negated and str_to_bool(self.dst_networks_negated):
            dst_negated = "NOT "
        if self.dst_services_negated and str_to_bool(self.dst_services_negated):
            srv_negated = "NOT "
        # TODO: Check if we also need to support source services here.
        if self.comment:
            comment_str = "COMMENT {}".format(self.comment)
        else:
            comment_str = ""
        return "ACTION {action} FROM {neg_src}{src} TO {neg_dst}{dst} SERVICE {neg_srv}{srv} {comment}".format(
                action=self.action, neg_src=src_negated, src=[str(src) for src in self.src_networks],
                neg_dst=dst_negated, dst=[str(dst) for dst in self.dst_networks], neg_srv=srv_negated,
                srv=[str(srv) for srv in self.dst_services], comment=comment_str)

    @property
    def tuple_header(self):
        if self.src_networks and self.dst_networks:
            return "Rule Number", "Source Negated", "Source", "Destination Negated", "Destination", "Services " \
                                                                                                    "Negated", \
                   "Services", "Action"
        else:
            return "VPN", "Application"

    def as_tuple(self):
        if self.src_networks and self.dst_networks:
            rule_content = (
                self.rule_number, self.src_networks_negated, ",".join([str(src) for src in self.src_networks]),
                self.dst_networks_negated, ",".join([str(dst) for dst in self.dst_networks]), self.dst_services_negated,
                ",".join([str(srv) for srv in self.dst_services]))
        else:
            rule_content = self.vpn, self.application
        return self.tuple_header, rule_content

    def is_enabled(self):
        return str_to_bool(self.disabled)


class Rule_Binding(XML_Object_Base):
    def __init__(self, acl, policy, default, rule_count, from_zone, to_zone, security_rule_count=None, uid=None,
                 direction=None):
        self.acl = acl
        self.policy = policy
        self.default = default
        self.rule_count = rule_count
        self.from_zone = from_zone
        self.to_zone = to_zone
        self.security_rule_count = security_rule_count
        self.uid = uid
        self.direction = direction
        super().__init__(xml_tags.Elements.BINDING)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        acl_node = get_xml_node(xml_node, xml_tags.Elements.ACL, True)
        policy_node = get_xml_node(xml_node, xml_tags.Elements.POLICY, True)
        if acl_node is not None:
            acl = Access_List.from_xml_node(acl_node)
        else:
            acl = None
        if policy_node is not None:
            policy = Policy.from_xml_node(policy_node)
        else:
            policy = None
        default = get_xml_text_value(xml_node, xml_tags.Elements.DEFAULT)
        from_zone_node = get_xml_node(xml_node, xml_tags.Elements.FROM_ZONE, True)
        if from_zone_node:
            from_zone = Rule_Binding_Zone.from_xml_node(from_zone_node)
        else:
            from_zone = None
        to_zone_node = get_xml_node(xml_node, xml_tags.Elements.TO_ZONE, True)
        if to_zone_node:
            to_zone = Rule_Binding_Zone.from_xml_node(to_zone_node)
        else:
            to_zone = None
        uid = get_xml_text_value(xml_node, xml_tags.Elements.UID)
        rule_count = get_xml_text_value(xml_node, xml_tags.Elements.RULE_COUNT)
        security_rule_count = get_xml_text_value(xml_node, xml_tags.Elements.SECURITY_RULE_COUNT)
        direction = get_xml_text_value(xml_node, xml_tags.Elements.DIRECTION)
        return cls(acl, policy, default, rule_count, from_zone, to_zone, security_rule_count, uid, direction)


class Destination_Network(Base_Object):
    def __init__(self, name, display_name, object_id=None):
        super().__init__(xml_tags.Elements.DST_NETWORK, name, display_name, object_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        object_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        return cls(name, display_name, object_id)


class Source_Network(Base_Object):
    def __init__(self, name, display_name, object_id=None):
        super().__init__(xml_tags.Elements.SRC_NETWORK, name, display_name, object_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        object_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        return cls(name, display_name, object_id)


class Destination_Service(Base_Object):
    def __init__(self, name, display_name, object_id=None):
        super().__init__(xml_tags.Elements.DST_SERVICE, name, display_name, object_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        object_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        return cls(name, display_name, object_id)


class RuleVPNOption(Base_Object):
    def __init__(self, name, display_name, object_id=None):
        super().__init__(xml_tags.Elements.VPN, name, display_name, object_id)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        object_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        return cls(name, display_name, object_id)


class Rule_Track(XML_Object_Base):
    NONE = "NONE"
    LOG = "LOG"

    def __init__(self, interval, level=None):
        self.level = level
        self.interval = interval
        super().__init__(xml_tags.Elements.TRACK)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        level = get_xml_text_value(xml_node, xml_tags.Elements.LEVEL)
        interval = get_xml_text_value(xml_node, xml_tags.Elements.INTERVAL)
        return cls(interval, level)

    def is_enabled(self):
        if self.level != Rule_Track.NONE:
            return True
        return False

    def is_logged(self):
        if self.level == Rule_Track.LOG:
            return True
        return False


class Access_List(XML_Object_Base):
    def __init__(self, is_global, interfaces=None, name=None):
        self.global_ = is_global
        self.interfaces = interfaces
        self.name = name
        super().__init__(xml_tags.Elements.ACL)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        global_ = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        interfaces = Interface.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.INTERFACES))
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        return cls(global_, interfaces, name)


class Policy(XML_Object_Base):
    def __init__(self, num_id, itg_id, itg, name, unique_active_in_itg=None):
        self.id = num_id
        self.itg = itg
        self.itg_id = itg_id
        self.name = name
        self.unique_active_in_itg = unique_active_in_itg
        super().__init__(xml_tags.Elements.POLICY)

    @classmethod
    def from_xml_node(cls, xml_node=None):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        num_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        itg = get_xml_text_value(xml_node, xml_tags.Elements.ITG)
        itg_id = get_xml_int_value(xml_node, xml_tags.Elements.ITG_ID)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        unique_active_in_itg = get_xml_text_value(xml_node, xml_tags.Elements.UNIQUE_ACTIVE_IN_ITG)
        return cls(num_id, itg_id, itg, name, unique_active_in_itg)

    def __repr__(self):
        return "Policy({id},{itg},{itg_id},'{name}',{unique_active_in_itg})".format(**self.__dict__)


class Interface_IP(XML_Object_Base, IPNetworkMixin):
    def __init__(self, ip, netmask=None, precedence=None, visibility=None):
        self.ip = ip
        self.netmask = netmask
        self.precedence = precedence
        self.visibility = visibility
        super().__init__(xml_tags.Elements.INTERFACE_IP)

    def _get_ip_network(self):
        """

        :rtype: netaddr.IPNetwork
        """
        return netaddr.IPNetwork(self.ip + "/" + self.netmask)

    def __str__(self):
        return "{}/{}".format(self.ip, netmask_to_cidr(self.netmask))

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        ip = get_xml_text_value(xml_node, xml_tags.Elements.IP)
        netmask = get_xml_text_value(xml_node, xml_tags.Elements.NETMASK)
        precedence = get_xml_text_value(xml_node, xml_tags.Elements.PRECEDENCE)
        visibility = get_xml_text_value(xml_node, xml_tags.Elements.VISIBILITY)
        return cls(ip, netmask, precedence, visibility)


class Interface(XSI_Object):
    def __init__(self, name, num_id, direction, device_id, acl_name, is_global, interface_ips=None):

        self.name = name
        self.id = num_id
        self.direction = direction
        self.device_id = device_id
        self.acl_name = acl_name
        self.interface_ips = interface_ips
        self.global_ = is_global
        super().__init__(xml_tags.Elements.INTERFACE, Attributes.INTERFACE_TYPE)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        num_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        direction = get_xml_text_value(xml_node, xml_tags.Elements.DIRECTION)
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        acl_name = get_xml_text_value(xml_node, xml_tags.Elements.ACL_NAME)
        global_ = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        interface_ips = XML_List(xml_tags.Elements.INTERFACE_IPS)
        for interface_ip_node in xml_node.iter(tag=xml_tags.Elements.INTERFACE_IP):
            interface_ips.append(Interface_IP.from_xml_node(interface_ip_node))
        return cls(name, num_id, direction, device_id, acl_name, global_, interface_ips)


class Topology_Interface(XML_Object_Base):
    def __init__(self, device_id, ip, mask, name, virtual_router, zone):
        self.device_id = device_id
        self.ip = ip
        self.mask = mask
        self.name = name
        self.virtual_router = virtual_router
        self.zone = zone
        super().__init__(xml_tags.Elements.INTERFACE)

    @classmethod
    def from_xml_node(cls, xml_node):
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        ip = get_xml_text_value(xml_node, xml_tags.Elements.IP)
        mask = get_xml_text_value(xml_node, xml_tags.Elements.MASK)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        virtual_router = get_xml_text_value(xml_node, xml_tags.Elements.VIRTUAL_ROUTER)
        zone = get_xml_text_value(xml_node, xml_tags.Elements.ZONE)
        return cls(device_id, ip, mask, name, virtual_router, zone)


class Services_List(XML_List):
    def __init__(self, services):
        """
        :type services: list[Single_Service]
        """
        self.services = services
        super().__init__(xml_tags.Elements.SERVICES, services)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        services = []
        for service_node in xml_node.iter(tag=xml_tags.Elements.SERVICE):
            if service_node.attrib[xml_tags.Attributes.XSI_NAMESPACE_TYPE] == xml_tags.Attributes.SERVICE_TYPE_SINGLE:
                services.append(Single_Service.from_xml_node(service_node))
            elif service_node.attrib[xml_tags.Attributes.XSI_NAMESPACE_TYPE] == xml_tags.Attributes.SERVICE_TYPE_GROUP:
                services.append(Group_Service.from_xml_node(service_node))
            else:
                raise ValueError("Unknown service type '{0}'.".format(
                        service_node.attrib[xml_tags.Attributes.XSI_NAMESPACE_TYPE]))
        return cls(services)


class Single_Service(Service):
    def __init__(self, service_id, display_name, is_global, name, service_type, protocol, port_min, port_max, negate,
                 comment):
        self.protocol = protocol
        self.min = port_min
        self.max = port_max
        self.negate = negate
        self.comment = comment
        super().__init__(xml_tags.Elements.SERVICE, service_id, display_name, is_global, name, service_type,
                         xml_tags.Attributes.SERVICE_TYPE_SINGLE)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        protocol = get_xml_int_value(xml_node, xml_tags.Elements.PROTOCOL)
        negate = get_xml_text_value(xml_node, xml_tags.Elements.NEGATE)
        port_min = get_xml_int_value(xml_node, xml_tags.Elements.MIN)
        port_max = get_xml_int_value(xml_node, xml_tags.Elements.MAX)
        comment = get_xml_text_value(xml_node, xml_tags.Elements.COMMENT)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        is_global = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        service_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        service_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        return cls(service_id, display_name, is_global, name, service_type, protocol, port_min, port_max, negate,
                   comment)

    def __str__(self):
        iana_protocols = get_iana_protocols()
        if self.min == self.max:
            return "{} {}".format(iana_protocols[int(self.protocol)], self.min)
        else:
            return "{} {}-{}".format(iana_protocols[int(self.protocol)], self.min, self.max)

    def as_service_type(self):
        if self.protocol is not None:
            if self.min == self.max:
                return Single_Service_Type(self.protocol, self.min)
            else:
                return Range_Service_Type(self.protocol, self.min, self.max)
        else:
            return Any_Service_Type()


class Group_Service(Service):
    def __init__(self, service_id, display_name, is_global, name, service_type, members):
        self.members = members
        super().__init__(xml_tags.Elements.SERVICE, service_id, display_name, is_global, name, service_type,
                         xml_tags.Attributes.SERVICE_TYPE_GROUP)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        is_global = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        service_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        service_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        members = XML_List(xml_tags.Elements.MEMBERS, [])
        for member_node in xml_node.iter(tag=xml_tags.Elements.MEMBER):
            member_id = get_xml_int_value(member_node, xml_tags.Elements.ID)
            member_display_name = get_xml_text_value(member_node, xml_tags.Elements.DISPLAY_NAME)
            member_name = get_xml_text_value(member_node, xml_tags.Elements.NAME)
            members.append(Base_Object(xml_tags.Elements.MEMBER, member_name, member_display_name, member_id))
        return cls(service_id, display_name, is_global, name, service_type, members)

    def __str__(self):
        spacer = 4 * " "
        if self.members:
            return "{}Members:\n{}{}".format(spacer, 2 * spacer, "\n{}".format(2 * spacer).join(
                    [member.display_name for member in self.members]))
        else:
            return "{}No members".format(spacer)

    def as_service_type(self):
        return Group_Service_Type(self.members)


class Network_Objects_List(XML_List):
    def __init__(self, network_objects):
        """
        :type network_objects: list[T <= Network_Object]
        """
        super().__init__(xml_tags.Elements.NETWORK_OBJECTS, network_objects)
        self.network_objects = network_objects

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        network_objects = []
        for network_object_node in xml_node.iter(tag=xml_tags.Elements.NETWORK_OBJECT):
            network_object_type = network_object_node.attrib[xml_tags.Attributes.XSI_NAMESPACE_TYPE]

            network_object = Basic_Network_Object.from_xml_node(network_object_node)
            if network_object_type == xml_tags.Attributes.NETWORK_OBJECT_TYPE_BASIC:
                network_objects.append(Basic_Network_Object.from_xml_node(network_object_node))
            elif network_object_type == xml_tags.Attributes.NETWORK_OBJECT_TYPE_RANGE:
                network_objects.append(Range_Network_Object.from_xml_node(network_object_node))
            elif network_object_type == xml_tags.Attributes.NETWORK_OBJECT_TYPE_HOST:
                network_objects.append(Host_Network_Object.from_xml_node(network_object_node))
            elif network_object_type == xml_tags.Attributes.NETWORK_OBJECT_TYPE_HOST_WITH_INTERFACES:
                network_objects.append(Host_With_Interfaces_Network_Object.from_xml_node(network_object_node))
            elif network_object_type == xml_tags.Attributes.NETWORK_OBJECT_TYPE_SUBNET:
                network_objects.append(Subnet_Network_Object.from_xml_node(network_object_node))
            elif network_object_type == xml_tags.Attributes.NETWORK_OBJECT_TYPE_GROUP:
                network_objects.append(Group_Network_Object.from_xml_node(network_object_node))
            elif network_object_type == xml_tags.Attributes.NETWORK_OBJECT_TYPE_CLOUD:
                network_objects.append(Cloud_Network_Object.from_xml_node(network_object_node))
            else:
                message = "Got unknown type '{}'".format(network_object_type)
                logger.error(message)
                raise ValueError(message)
        return cls(network_objects)


class Basic_Network_Object(Network_Object):
    def __init__(self, display_name, is_global, object_id, name, object_type, ip, device_id, comment, implicit,
                 class_name=None):

        self.set_attrib(xml_tags.Attributes.XSI_TYPE, xml_tags.Attributes.NETWORK_OBJECT_TYPE_BASIC)
        self.ip = ip
        self.global_ = is_global
        super().__init__(xml_tags.Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         device_id, comment, implicit, class_name)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        is_global = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        ip = get_xml_text_value(xml_node, xml_tags.Elements.IP)
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        comment = get_xml_text_value(xml_node, xml_tags.Elements.COMMENT)
        implicit = get_xml_text_value(xml_node, xml_tags.Elements.IMPLICIT)
        class_name = get_xml_text_value(xml_node, xml_tags.Elements.CLASS_NAME)
        return cls(display_name, is_global, object_id, name, object_type, ip, device_id, comment, implicit, class_name)

    def __str__(self):
        if self.ip is not None:
            return str(self.ip)
        else:
            return "Any"

    def as_netaddr_obj(self):
        if self.ip is not None:
            return netaddr.IPNetwork(self.ip)
        else:
            return netaddr.IPNetwork("0.0.0.0/0")


class Range_Network_Object(Network_Object):
    def __init__(self, display_name, is_global, object_id, name, object_type, first_ip, last_ip, device_id, comment,
                 implicit):
        self.set_attrib(xml_tags.Attributes.XSI_TYPE, xml_tags.Attributes.NETWORK_OBJECT_TYPE_RANGE)
        self.first_ip = first_ip
        self.last_ip = last_ip
        super().__init__(xml_tags.Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         device_id, comment, implicit)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        is_global = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        first_ip = get_xml_text_value(xml_node, xml_tags.Elements.FIRST_IP)
        last_ip = get_xml_text_value(xml_node, xml_tags.Elements.LAST_IP)
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        comment = get_xml_text_value(xml_node, xml_tags.Elements.COMMENT)
        implicit = get_xml_text_value(xml_node, xml_tags.Elements.IMPLICIT)
        return cls(display_name, is_global, object_id, name, object_type, first_ip, last_ip, device_id, comment,
                   implicit)

    def __str__(self):
        return "{}-{}".format(self.first_ip, self.last_ip)

    def as_netaddr_obj(self):
        return netaddr.IPRange(self.first_ip, self.last_ip, flags=netaddr.ZEROFILL)


class Host_Network_Object(Network_Object):
    def __init__(self, display_name, is_global, object_id, name, object_type, ip, device_id, comment, implicit):
        self.ip = ip
        self.set_attrib(xml_tags.Attributes.XSI_TYPE, xml_tags.Attributes.NETWORK_OBJECT_TYPE_HOST)
        super().__init__(xml_tags.Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         device_id, comment, implicit)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        is_global = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        ip = get_xml_text_value(xml_node, xml_tags.Elements.IP)
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        comment = get_xml_text_value(xml_node, xml_tags.Elements.COMMENT)
        implicit = get_xml_text_value(xml_node, xml_tags.Elements.IMPLICIT)
        return cls(display_name, is_global, object_id, name, object_type, ip, device_id, comment, implicit)

    def __str__(self):
        return self.ip

    def as_netaddr_obj(self):
        return netaddr.IPNetwork(self.ip)


class Host_With_Interfaces_Network_Object(Host_Network_Object):
    def __init__(self, display_name, is_global, object_id, name, object_type, ip, class_name, interfaces, device_id,
                 comment, implicit):
        self.set_attrib(xml_tags.Attributes.XSI_TYPE, xml_tags.Attributes.NETWORK_OBJECT_TYPE_HOST_WITH_INTERFACES)
        self.class_name = class_name
        self.interfaces = interfaces
        super().__init__(display_name, is_global, object_id, name, object_type, ip, device_id, comment, implicit)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        is_global = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        ip = get_xml_text_value(xml_node, xml_tags.Elements.IP)
        class_name = get_xml_text_value(xml_node, xml_tags.Elements.CLASS_NAME)
        interfaces = XML_List.from_xml_node_by_tags(xml_node, xml_tags.Elements.INTERFACES, xml_tags.Elements.INTERFACE,
                                                    Host_Interface, optional=True)
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        comment = get_xml_text_value(xml_node, xml_tags.Elements.COMMENT)
        implicit = get_xml_text_value(xml_node, xml_tags.Elements.IMPLICIT)
        return cls(display_name, is_global, object_id, name, object_type, ip, class_name, interfaces, device_id,
                   comment, implicit)

    def __str__(self):
        return self.ip


class Host_Interface(XML_Object_Base):
    def __init__(self, ip=None, mask=None, name=None, interface_ips=None):
        self.set_attrib(xml_tags.Attributes.XSI_TYPE, xml_tags.Attributes.NETWORK_OBJECT_TYPE_HOST)
        self.name = name
        if interface_ips:
            self.interface_ips = interface_ips
        else:
            self.ip = ip
            self.mask = mask
        super().__init__(xml_tags.Elements.INTERFACE)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        try:
            interface_ips = XML_List.from_xml_node_by_tags(xml_node, xml_tags.Elements.INTERFACE_IPS,
                                                           xml_tags.Elements.INTERFACE_IP, Interface_IP)
            return cls(name=name, interface_ips=interface_ips)
        except ValueError:
            ip = get_xml_text_value(xml_node, xml_tags.Elements.IP)
            mask = get_xml_text_value(xml_node, xml_tags.Elements.MASK)

            return cls(ip=ip, mask=mask, name=name)


class Subnet_Network_Object(Network_Object):
    def __init__(self, display_name, is_global, object_id, name, object_type, ip, netmask, device_id, comment,
                 implicit):
        self.netmask = netmask
        self.ip = ip
        self.set_attrib(xml_tags.Attributes.XSI_TYPE, xml_tags.Attributes.NETWORK_OBJECT_TYPE_SUBNET)
        super().__init__(xml_tags.Elements.NETWORK_OBJECT, display_name, is_global, object_id, name, object_type,
                         device_id, comment, implicit)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        object_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        is_global = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        object_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        netmask = get_xml_text_value(xml_node, xml_tags.Elements.NETMASK)
        ip = get_xml_text_value(xml_node, xml_tags.Elements.IP)
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        comment = get_xml_text_value(xml_node, xml_tags.Elements.COMMENT)
        implicit = get_xml_text_value(xml_node, xml_tags.Elements.IMPLICIT)
        return cls(display_name, is_global, object_id, name, object_type, ip, netmask, device_id, comment, implicit)

    def __str__(self):
        return "{}/{}".format(self.ip, self.netmask)

    def as_netaddr_obj(self):
        return netaddr.IPNetwork("{}/{}".format(self.ip, '.'.join(str(int(o)) for o in self.netmask.split('.'))))


class Base_Network_Object(Network_Object):
    def __init__(self, display_name, is_global, connection_id, name, service_type, members, uid, device_id, comment,
                 implicit):
        self.members = members
        self.uid = uid
        super().__init__(xml_tags.Elements.NETWORK_OBJECT, display_name, is_global, connection_id, name, service_type,
                         device_id, comment, implicit)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        display_name = get_xml_text_value(xml_node, xml_tags.Elements.DISPLAY_NAME)
        is_global = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        connection_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        uid = get_xml_text_value(xml_node, xml_tags.Elements.UID)
        service_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        members = XML_List(xml_tags.Elements.MEMBERS, [])
        for member_node in xml_node.iter(tag=xml_tags.Elements.MEMBER):
            member_id = get_xml_int_value(member_node, xml_tags.Elements.ID)
            member_display_name = get_xml_text_value(member_node, xml_tags.Elements.DISPLAY_NAME)
            member_name = get_xml_text_value(member_node, xml_tags.Elements.NAME)
            members.append(Base_Object(xml_tags.Elements.MEMBER, member_name, member_display_name, member_id))
        device_id = get_xml_int_value(xml_node, xml_tags.Elements.DEVICE_ID)
        comment = get_xml_text_value(xml_node, xml_tags.Elements.COMMENT)
        implicit = get_xml_text_value(xml_node, xml_tags.Elements.IMPLICIT)
        return cls(display_name, is_global, connection_id, name, service_type, members, uid, device_id, comment,
                   implicit)

    def __str__(self):
        spacer = 4 * " "
        if self.members:
            return "{}Members:\n{}{}".format(spacer, 2 * spacer, "\n{}".format(2 * spacer).join(
                    [member.display_name for member in self.members]))
        else:
            return "{}No members".format(spacer)


class Group_Network_Object(Base_Network_Object):
    def __init__(self, display_name, is_global, connection_id, name, service_type, members, uid, device_id, comment,
                 implicit):
        self.set_attrib(xml_tags.Attributes.XSI_TYPE, xml_tags.Attributes.NETWORK_OBJECT_TYPE_GROUP)
        super().__init__(display_name, is_global, connection_id, name, service_type, members, uid, device_id, comment,
                         implicit)


class Cloud_Network_Object(Base_Network_Object):
    def __init__(self, display_name, is_global, connection_id, name, service_type, members, uid, device_id, comment,
                 implicit):
        self.set_attrib(xml_tags.Attributes.XSI_TYPE, xml_tags.Attributes.NETWORK_OBJECT_TYPE_CLOUD)
        super().__init__(display_name, is_global, connection_id, name, service_type, members, uid, device_id, comment,
                         implicit)


class Policy_Analysis_Query_Result(XML_Object_Base):
    def __init__(self, devices_and_bindings):
        self.devices_and_bindings = devices_and_bindings
        super().__init__(xml_tags.Elements.POLICY_ANALYSIS_QUERY_RESULT)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        devices_and_bindings_node = get_xml_node(xml_node, xml_tags.Elements.DEVICES_AND_BINDINGS)
        devices_and_bindings = Devices_And_Bindings_List.from_xml_node(devices_and_bindings_node)
        return cls(devices_and_bindings)


class Devices_And_Bindings_List(XML_List):
    def __init__(self, items):
        """
        :type items: list[Device_And_Bindings]
        """
        self.items = items
        super().__init__(xml_tags.Elements.DEVICES_AND_BINDINGS, items)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        items = []
        for device_and_bindings_node in xml_node:
            items.append(Device_And_Bindings.from_xml_node(device_and_bindings_node))
        return cls(items)


class Bindings_And_Rules_List(XML_List):
    def __init__(self, items):
        """
        :type items: list[Binding_And_Rules]
        """
        self.items = items
        super().__init__(xml_tags.Elements.BINDINGS_AND_RULES, items)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        items = []
        for device_and_bindings_node in xml_node:
            items.append(Binding_And_Rules.from_xml_node(device_and_bindings_node))
        return cls(items)


class Binding_And_Rules(XML_Object_Base):
    def __init__(self, bindings, rules):
        self.bindings = bindings
        self.rules = rules
        super().__init__(xml_tags.Elements.BINDING_AND_RULES)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        rules = []
        rules_node = get_xml_node(xml_node, xml_tags.Elements.RULES)
        for rule_node in rules_node.iter(tag=xml_tags.Elements.RULE):
            rules.append(Rule.from_xml_node(rule_node))
        binding_node = get_xml_node(xml_node, xml_tags.Elements.BINDING)
        binding = Rule_Binding.from_xml_node(binding_node)
        return cls(binding, rules)


class Device_And_Bindings(XML_Object_Base):
    def __init__(self, device, bindings_and_rules):
        self.device = device
        self.bindings_and_rules = bindings_and_rules
        super().__init__(xml_tags.Elements.DEVICE_AND_BINDINGS)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        device_node = get_xml_node(xml_node, xml_tags.Elements.DEVICE)
        device = Device.from_xml_node(device_node)
        bindings_and_rules_node = get_xml_node(xml_node, xml_tags.Elements.BINDINGS_AND_RULES)
        bindings_and_rules = Bindings_And_Rules_List.from_xml_node(bindings_and_rules_node)
        return cls(device, bindings_and_rules)


class SecureApp_Application(XML_Object_Base):
    def __init__(self, app_name, app_owner):
        """
        Initialize the object from parameters.
        :param app_name: The name of the application.
        :type app_name: str
        :param app_owner: The application owner.
        :type app_owner: str
        """
        self.app_name = app_name
        self.app_owner = app_owner
        super().__init__(xml_tags.Elements.SECURE_APP_APPLICATION)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        app_owner = get_xml_text_value(xml_node, xml_tags.Elements.APP_OWNER)
        app_name = get_xml_text_value(xml_node, xml_tags.Elements.APP_NAME)
        return cls(app_name, app_owner)


class Interfaces_List(XML_List):
    def __init__(self, interfaces):
        """
        :type interfaces: list[Interface]
        """
        super().__init__(xml_tags.Elements.INTERFACES, interfaces)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        interfaces = []
        for interface_node in xml_node.iter(tag=xml_tags.Elements.INTERFACE):
            interfaces.append(Interface.from_xml_node(interface_node))
        return cls(interfaces)


class Topology_Interfaces_List(XML_List):
    def __init__(self, interfaces):
        """
        :type interfaces: list[Topology_Interface]
        """
        super().__init__(xml_tags.Elements.INTERFACES, interfaces)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        interfaces = []
        for interface_node in xml_node.iter(tag=xml_tags.Elements.INTERFACE):
            interfaces.append(Topology_Interface.from_xml_node(interface_node))
        return cls(interfaces)


class Change_Rules(XML_Object_Base):
    def __init__(self, xml_tag, new_rule, old_rules, old_rules_violated_traffic):
        """
        :type xml_tag: str
        :type new_rule: Rule
        :type old_rules: XML_List
        :type old_rules_violated_traffic: OldRulesViolatedTraffic
        """
        super().__init__(xml_tag)
        self.new_rule = new_rule
        self.old_rules = old_rules
        self.old_rules_violated_traffic = old_rules_violated_traffic

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        # TODO: Rule has an incorrect tag of "rule", should be "new_rule" here.
        new_rule = Rule.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.NEW_RULE))
        new_rule._xml_tag = xml_tags.Elements.NEW_RULE
        old_rules = XML_List.from_xml_node_by_tags(xml_node, xml_tags.Elements.OLD_RULES, xml_tags.Elements.OLD_RULE,
                                                   Rule)
        old_rules_violated_traffic_node = get_xml_node(xml_node, xml_tags.Elements.OLD_RULES_VIOLATED_TRAFFIC, True)
        if old_rules_violated_traffic_node is not None:
            old_rules_violated_traffic = OldRulesViolatedTraffic.from_xml_node(old_rules_violated_traffic_node)
        else:
            old_rules_violated_traffic = None
        return cls(new_rule, old_rules, old_rules_violated_traffic)


class OldRulesViolatedTraffic(XML_Object_Base):
    def __init__(self, rule_violated_traffic):
        """
        :type rule_violated_traffic: XML_List[TrafficRange]
        """
        super().__init__(xml_tags.Elements.OLD_RULES_VIOLATED_TRAFFIC)
        self.rule_violated_traffic = rule_violated_traffic

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        rule_violated_traffic_node = get_xml_node(xml_node, xml_tags.Elements.RULE_VIOLATED_TRAFFIC)
        rule_violated_traffic = RuleViolatedTraffic.from_xml_node(rule_violated_traffic_node)
        return cls(rule_violated_traffic)


class RuleViolatedTraffic(XML_Object_Base):
    def __init__(self, violated_traffic):
        super().__init__(xml_tags.Elements.RULE_VIOLATED_TRAFFIC)
        self.violated_traffic = violated_traffic

    @classmethod
    def from_xml_node(cls, xml_node):
        violated_traffic = ViolatedTraffic.from_xml_node(xml_node)
        return cls(violated_traffic)


class ViolatedTraffic(XML_List):
    def __init__(self, traffic_ranges):
        super().__init__(xml_tags.Elements.VIOLATED_TRAFFIC, traffic_ranges)

    @classmethod
    def from_xml_node(cls, xml_node):
        traffic_ranges = XML_List.from_xml_node_by_tags(xml_node, xml_tags.Elements.VIOLATED_TRAFFIC,
                                                        xml_tags.Elements.TRAFFIC_RANGE, TrafficRange)
        return cls(traffic_ranges)


class Change_Accepting_Rules(Change_Rules):
    def __init__(self, new_rule, old_rules, old_rules_violated_traffic):
        super().__init__(xml_tags.Elements.ACCEPTINGRULESDTO, new_rule, old_rules, old_rules_violated_traffic)


class Change_Blocking_Rules(Change_Rules):
    def __init__(self, new_rule, old_rules, old_rules_violated_traffic):
        super().__init__(xml_tags.Elements.BLOCKINGRULESDTO, new_rule, old_rules, old_rules_violated_traffic)


class Change_Authorization(XML_Object_Base):
    UNAUTHORIZED = "unauthorized"
    AUTHORIZED = "authorized"

    def __init__(self, status, new_revision, old_revision, change_auth_bindings):
        """
        :type status: str
        :type new_revision: Device_Revision
        :type old_revision: Device_Revision
        :type change_auth_bindings: ChangeAuthorizationBindings
        """
        super().__init__(xml_tags.Elements.CHANGE_AUTHORIZATION)
        self.status = status
        self.new_revision = new_revision
        self.old_revision = old_revision
        self.change_authorization_bindings = change_auth_bindings

    @classmethod
    def from_xml_node(cls, xml_node=None):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        status = get_xml_text_value(xml_node, xml_tags.Elements.STATUS)
        new_revision_node = get_xml_node(xml_node, xml_tags.Elements.NEW_REVISION)
        new_revision = Device_Revision.from_xml_node(new_revision_node)
        old_revision_node = get_xml_node(xml_node, xml_tags.Elements.OLD_REVISION)
        old_revision = Device_Revision.from_xml_node(old_revision_node)
        change_auth_bindings_node = get_xml_node(xml_node, xml_tags.Elements.CHANGE_AUTHORIZATION_BINDINGS)
        change_auth_bindings = ChangeAuthorizationBindings.from_xml_node(change_auth_bindings_node)
        return cls(status, new_revision, old_revision, change_auth_bindings)

    def is_authorized(self):
        return self.status.lower() == self.AUTHORIZED.lower()


class ChangeAuthorizationBinding(XML_Object_Base):
    def __init__(self, binding, unauthorized_opened_access, unauthorized_closed_access):
        """
        :type binding: Rule_Binding
        :type unauthorized_opened_access: XML_List[Change_Accepting_Rules]
        :type unauthorized_closed_access: XML_List[Change_Blocking_Rules]
        """
        super().__init__(xml_tags.Elements.CHANGE_AUTHORIZATION_BINDING)
        self.binding = binding
        self.unauthorized_opened_access = unauthorized_opened_access
        self.unauthorized_closed_access = unauthorized_closed_access

    @classmethod
    def from_xml_node(cls, xml_node):
        unauthorized_opened_access = XML_List.from_xml_node_by_tags(xml_node,
                                                                    xml_tags.Elements.UNAUTHORIZED_OPENED_ACCESS,
                                                                    xml_tags.Elements.ACCEPTINGRULESDTO,
                                                                    Change_Accepting_Rules)

        unauthorized_closed_access = XML_List.from_xml_node_by_tags(xml_node,
                                                                    xml_tags.Elements.UNAUTHORIZED_CLOSED_ACCESS,
                                                                    xml_tags.Elements.BLOCKINGRULESDTO,
                                                                    Change_Blocking_Rules)
        binding_node = get_xml_node(xml_node, xml_tags.Elements.BINDING, True)
        if binding_node:
            binding = Rule_Binding.from_xml_node(binding_node)
        else:
            binding = None
        return cls(binding, unauthorized_opened_access, unauthorized_closed_access)


class ChangeAuthorizationBindings(XML_List):
    def __init__(self, change_auth_bindings):
        """
        :type change_auth_bindings: list[ChangeAuthorizationBinding]
        """
        super().__init__(xml_tags.Elements.CHANGE_AUTHORIZATION_BINDINGS, change_auth_bindings)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        change_auth_bindings = []
        for change_auth_binding_node in xml_node.iter(tag=xml_tags.Elements.CHANGE_AUTHORIZATION_BINDING):
            change_auth_bindings.append(ChangeAuthorizationBinding.from_xml_node(change_auth_binding_node))
        return cls(change_auth_bindings)


class Rule_Binding_Zone(XML_Object_Base):
    def __init__(self, xml_tag, name, is_global, zone_id):
        super().__init__(xml_tag)
        self.id = zone_id
        self.global_ = is_global
        self.name = name

    @classmethod
    def from_xml_node(cls, xml_node):
        zone_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        name = get_xml_text_value(xml_node, xml_tags.Elements.NAME)
        is_global = get_xml_text_value(xml_node, xml_tags.Elements.GLOBAL)
        return cls(xml_node.tag, name, is_global, zone_id)


class SecurityRequirement(XML_Object_Base):
    def __init__(self, policy_control_name, access_type, from_zone, to_zone, rule_properties, flow, allowed_services,
                 blocked_services):
        super().__init__(xml_tags.Elements.SECURITY_REQUIREMENT)
        self.policy_control_name = policy_control_name
        self.access_type = access_type
        self.from_zone = from_zone
        self.to_zone = to_zone
        self.rule_properties = rule_properties
        self.flow = flow
        self.allowed_services = allowed_services
        self.blocked_services = blocked_services

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        policy_control_name = get_xml_text_value(xml_node, xml_tags.Elements.POLICY_CONTROL_NAME)
        access_type = get_xml_text_value(xml_node, xml_tags.Elements.ACCESS_TYPE)
        from_zone = get_xml_text_value(xml_node, xml_tags.Elements.FROM_ZONE)
        to_zone = get_xml_text_value(xml_node, xml_tags.Elements.TO_ZONE)

        rule_properties_node = get_xml_node(xml_node, xml_tags.Elements.RULE_PROPERTIES, optional=True)
        if rule_properties_node is None:
            rule_properties = None
        else:
            rule_properties = RuleProperties.from_xml_node(rule_properties_node)

        flow_node = get_xml_node(xml_node, xml_tags.Elements.FLOW, optional=True)
        if flow_node is None:
            flow = None
        else:
            flow = Flow.from_xml_node(flow_node)

        allowed_services_node = get_xml_node(xml_node, xml_tags.Elements.ALLOWED_SERVICES, optional=True)
        if allowed_services_node is None:
            allowed_services = None
        else:
            allowed_services = AllowedServices.from_xml_node(allowed_services_node)

        blocked_services_node = get_xml_node(xml_node, xml_tags.Elements.BLOCKED_SERVICES, optional=True)
        if blocked_services_node is None:
            blocked_services = None
        else:
            blocked_services = BlockedServices.from_xml_node(blocked_services_node)

        return cls(policy_control_name, access_type, from_zone, to_zone, rule_properties, flow, allowed_services,
                   blocked_services)


class AllowedOrBlockedServicesBase(XML_Object_Base):
    """
    Base class for either the "allowed services" or "blocked services" nodes of the "security_requirement" node
    (violations XML)
    """

    def __init__(self, id_, exclude_any, is_any, negate, service_items, xml_tag):
        super().__init__(xml_tag)
        self.id = id_
        self.exclude_any = exclude_any
        self.is_any = is_any
        self.negate = negate
        self.service_items = service_items


class AllowedServices(AllowedOrBlockedServicesBase):
    """
    Parses the "allowed services" node of the "security_requirement" node in the violations XML
    """

    def __init__(self, id_, exclude_any, is_any, negate, service_items):
        super().__init__(id_, exclude_any, is_any, negate, service_items, xml_tags.Elements.ALLOWED_SERVICES)

    @classmethod
    def from_xml_node(cls, xml_node):
        allowed_services_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        exclude_any = get_xml_text_value(xml_node, xml_tags.Elements.EXCLUDE_ANY)
        is_any = get_xml_text_value(xml_node, xml_tags.Elements.IS_ANY)
        negate = get_xml_text_value(xml_node, xml_tags.Elements.NEGATE)
        service_items = ServiceItemsList.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.SERVICE_ITEMS))
        return cls(allowed_services_id, exclude_any, is_any, negate, service_items)


class BlockedServices(AllowedOrBlockedServicesBase):
    """
    Parses the "blocked services" node of the "security_requirement" node in the violations XML
    """

    def __init__(self, id_, exclude_any, is_any, negate, service_items):
        super().__init__(id_, exclude_any, is_any, negate, service_items, xml_tags.Elements.BLOCKED_SERVICES)

    @classmethod
    def from_xml_node(cls, xml_node):
        blocked_services_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        exclude_any = get_xml_text_value(xml_node, xml_tags.Elements.EXCLUDE_ANY)
        is_any = get_xml_text_value(xml_node, xml_tags.Elements.IS_ANY)
        negate = get_xml_text_value(xml_node, xml_tags.Elements.NEGATE)
        service_items = ServiceItemsList.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.SERVICE_ITEMS))
        return cls(blocked_services_id, exclude_any, is_any, negate, service_items)


class ServiceItem(XML_Object_Base):
    def __init__(self, id_, port, protocol):
        super().__init__(xml_tags.Elements.SERVICE_ITEM)
        self.id = id_
        self.port = port
        self.protocol = protocol

    @classmethod
    def from_xml_node(cls, xml_node):
        service_item_id = get_xml_int_value(xml_node, xml_tags.Elements.ID)
        port = get_xml_text_value(xml_node, xml_tags.Elements.PORT)
        protocol = get_xml_text_value(xml_node, xml_tags.Elements.PROTOCOL)
        return cls(service_item_id, port, protocol)


class ServiceItemsList(XML_List):
    def __init__(self, service_items):
        self.service_items = service_items
        super().__init__(xml_tags.Elements.SERVICE_ITEMS, service_items)

    @classmethod
    def from_xml_node(cls, xml_node):
        services_items = []
        for service_item in xml_node.iter(tag=xml_tags.Elements.SERVICE_ITEM):
            service_item_obj = ServiceItem.from_xml_node(service_item)
            if service_item_obj is not None:
                services_items.append(service_item_obj)
        return cls(services_items)


class Flow(XML_Object_Base):
    """
    Parses the "flow" node of the "security_requirement" node. Represents the configured flow in the USP matrix.
    """

    def __init__(self, description, property_type):
        super().__init__(xml_tags.Elements.FLOW)
        self.description = description
        self.type = property_type

    @classmethod
    def from_xml_node(cls, xml_node):
        description = get_xml_text_value(xml_node, xml_tags.Elements.DESCRIPTION)
        property_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        return cls(description, property_type)


class RuleProperty(XML_Object_Base):
    """
    Parses a "rule property" node of the "security_requirement" node.
    """

    def __init__(self, description, property_type, payload):
        super().__init__(xml_tags.Elements.RULE_PROPERTY)
        self.description = description
        self.type = property_type
        self.payload = payload

    @classmethod
    def from_xml_node(cls, xml_node):
        description = get_xml_text_value(xml_node, xml_tags.Elements.DESCRIPTION)
        property_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        payload = get_xml_text_value(xml_node, xml_tags.Elements.PAYLOAD)
        return cls(description, property_type, payload)


class RuleProperties(XML_List):
    def __init__(self, rule_properties_list):
        self.rule_properties_list = rule_properties_list
        super().__init__(xml_tags.Elements.RULE_PROPERTY, rule_properties_list)

    @classmethod
    def from_xml_node(cls, xml_node):
        rule_properties_list = []
        for rule_property in xml_node.iter(tag=xml_tags.Elements.RULE_PROPERTY):
            rule_property_obj = RuleProperty.from_xml_node(rule_property)
            if rule_property_obj is not None:
                rule_properties_list.append(rule_property_obj)
        return cls(rule_properties_list)


class RulePropertyViolation(XML_Object_Base):
    def __init__(self, description, property_type):
        super().__init__(xml_tags.Elements.RULE_PROPERTY_VIOLATION)
        self.description = description
        self.type = property_type

    @classmethod
    def from_xml_node(cls, xml_node):
        description = get_xml_text_value(xml_node, xml_tags.Elements.DESCRIPTION)
        property_type = get_xml_text_value(xml_node, xml_tags.Elements.TYPE)
        return cls(description, property_type)


class RulePropertyViolations(XML_List):
    def __init__(self, rule_properties_violations_list):
        self.rule_properties_violations_list = rule_properties_violations_list
        super().__init__(xml_tags.Elements.RULE_PROPERTY_VIOLATION, rule_properties_violations_list)

    @classmethod
    def from_xml_node(cls, xml_node):
        rule_properties_violations_list = []
        for rule_property_violation in xml_node.iter(tag=xml_tags.Elements.RULE_PROPERTY_VIOLATION):
            rule_property_violation_obj = RulePropertyViolation.from_xml_node(rule_property_violation)
            if rule_property_violation_obj is not None:
                rule_properties_violations_list.append(rule_property_violation_obj)
        return cls(rule_properties_violations_list)


class ViolatingObject(XML_Object_Base):
    def __init__(self, path):
        super().__init__(xml_tags.Elements.VIOLATING_OBJECT)
        self.path = path

    @classmethod
    def from_xml_node(cls, xml_node):
        path = get_xml_text_value(xml_node, xml_tags.Elements.PATH)
        return cls(path)


class FlowViolationsBase(XML_Object_Base):
    """
    Base class for the the "flow source violations" and "flow destination violations" nodes of the
    "violation" node.
    """

    def __init__(self, xml_tag, negated, violating_objects):
        super().__init__(xml_tag)
        self.negated = negated
        self.violating_object = violating_objects

    @classmethod
    def get_negated_and_violating_rules(cls, xml_node):
        negated = get_xml_text_value(xml_node, xml_tags.Elements.NEGATED)
        violating_objects = []
        for violating_object_node in xml_node.iter(tag=xml_tags.Elements.VIOLATING_OBJECT):
            violating_object = ViolatingObject.from_xml_node(violating_object_node)
            if violating_object is not None:
                violating_objects.append(violating_object)
        return negated, violating_objects


class FlowSourceViolations(FlowViolationsBase):
    def __init__(self, negated, violating_objects):
        super().__init__(xml_tags.Elements.FLOW_SOURCE_VIOLATIONS, negated, violating_objects)

    @classmethod
    def from_xml_node(cls, xml_node):
        negated, violating_objects = super().get_negated_and_violating_rules(xml_node)
        return cls(negated, violating_objects)


class FlowDestinationViolations(FlowViolationsBase):
    def __init__(self, negated, violating_objects):
        super().__init__(xml_tags.Elements.FLOW_DESTINATION_VIOLATIONS, negated, violating_objects)

    @classmethod
    def from_xml_node(cls, xml_node):
        negated, violating_objects = super().get_negated_and_violating_rules(xml_node)
        return cls(negated, violating_objects)


class Path(Flat_XML_Object_Base):
    def __init__(self, path):
        super().__init__(xml_tag=xml_tags.Elements.PATH, content=path)

    @classmethod
    def from_xml_node(cls, xml_node):
        return cls(get_xml_text_value(xml_node, xml_tags.Elements.PATH))


class ViolationNetwork(XML_Object_Base):
    def __init__(self, paths):
        super().__init__(xml_tags.Elements.NETWORK)
        self.paths = paths

    @classmethod
    def from_xml_node(cls, xml_node):
        paths = []
        for path in xml_node.iter(tag=xml_tags.Elements.NETWORK):
            paths.append(Path.from_xml_node(path))
        return cls(paths)

    def get_all_paths_as_string(self):
        return '-'.join(path.content for path in self.paths)


class SourcesInZone(XML_Object_Base):
    def __init__(self, zone_name, networks):
        super().__init__(xml_tags.Elements.SOURCES_IN_ZONE)
        self.zone_name = zone_name
        self.networks = networks

    @classmethod
    def from_xml_node(cls, xml_node):
        sources_in_zone_node = get_xml_node(xml_node, xml_tags.Elements.SOURCES_IN_ZONE, optional=True)
        if sources_in_zone_node is None:
            return None
        zone_name = get_xml_text_value(sources_in_zone_node, xml_tags.Elements.ZONE_NAME)
        networks = XML_List.from_xml_node_by_tags(sources_in_zone_node, xml_tags.Elements.NETWORKS,
                                                  xml_tags.Elements.NETWORK, ViolationNetwork)
        return cls(zone_name, networks)


class DestinationsInZone(XML_Object_Base):
    def __init__(self, zone_name, networks):
        super().__init__(xml_tags.Elements.DESTINATIONS_IN_ZONE)
        self.zone_name = zone_name
        self.networks = networks

    @classmethod
    def from_xml_node(cls, xml_node):
        destinations_in_zone_node = get_xml_node(xml_node, xml_tags.Elements.DESTINATIONS_IN_ZONE, optional=True)
        if destinations_in_zone_node is None:
            return None
        zone_name = get_xml_text_value(destinations_in_zone_node, xml_tags.Elements.ZONE_NAME)
        networks = XML_List.from_xml_node_by_tags(destinations_in_zone_node, xml_tags.Elements.NETWORKS,
                                                  xml_tags.Elements.NETWORK, ViolationNetwork)
        return cls(zone_name, networks)


class ViolatingService(XML_Object_Base):
    def __init__(self, paths):
        self.paths = paths
        super().__init__(xml_tags.Elements.VIOLATING_SERVICE)

    @classmethod
    def from_xml_node(cls, xml_node):
        paths = []
        for path in xml_node.iter(tag=xml_tags.Elements.VIOLATING_SERVICE):
            paths.append(Path.from_xml_node(path))
        return cls(paths)

    def get_all_paths_as_string(self):
        return '-'.join(path.content for path in self.paths)


class ViolatingServices(XML_Object_Base):
    def __init__(self, negated, services):
        self.negated = negated
        self.services = services
        super().__init__(xml_tags.Elements.VIOLATING_SERVICES)

    @classmethod
    def from_xml_node(cls, xml_node):
        violating_services_node = get_xml_node(xml_node, xml_tags.Elements.VIOLATING_SERVICES)
        if violating_services_node is None:
            return None
        negated = get_xml_text_value(violating_services_node, xml_tags.Elements.NEGATED)
        services_list = []
        for violating_service in violating_services_node.iter(tag=xml_tags.Elements.VIOLATING_SERVICE):
            services_list.append(ViolatingService.from_xml_node(violating_service))
        return cls(negated, services_list)


class Violation(XML_Object_Base):
    # TODO: expend the Class to Any type of violation
    def __init__(self, severity, security_requirement, rule_properties_violations, flow_source_violations,
                 flow_destination_violations, sources_in_zone=None, destinations_in_zone=None, violating_services=None):
        super().__init__(xml_tags.Elements.VIOLATION)
        self.security_requirement = security_requirement
        self.severity = severity
        self.rule_properties_violations = rule_properties_violations
        self.flow_source_violations = flow_source_violations
        self.flow_destination_violations = flow_destination_violations
        self.sources_in_zone = sources_in_zone
        self.destinations_in_zone = destinations_in_zone
        self.violating_services = violating_services

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        severity = get_xml_text_value(xml_node, xml_tags.Elements.SEVERITY)
        # TODO: currently it parses only single type of requirement, should be done for all types and done by dict
        try:
            xpath = '{}[@{}="{}"]'.format(xml_tags.Elements.SECURITY_REQUIREMENT,
                                          xml_tags.Attributes.XSI_NAMESPACE_TYPE,
                                          xml_tags.Attributes.SECURITY_REQUIREMENT_TYPE_MATRIX)
            security_requirement_node = xml_node.findall(xpath)[0]
        except (ValueError, IndexError):
            return None
        else:
            security_requirement = SecurityRequirement.from_xml_node(security_requirement_node)

            rule_properties_violations_node = get_xml_node(xml_node, xml_tags.Elements.RULE_PROPERTIES_VIOLATIONS,
                                                           optional=True)
            if rule_properties_violations_node is None:
                rule_properties_violations = None
            else:
                rule_properties_violations = RulePropertyViolations.from_xml_node(rule_properties_violations_node)

            flow_source_violations_node = get_xml_node(xml_node, xml_tags.Elements.FLOW_SOURCE_VIOLATIONS,
                                                       optional=True)
            if flow_source_violations_node is None:
                flow_source_violations = None
            else:
                flow_source_violations = FlowSourceViolations.from_xml_node(flow_source_violations_node)

            flow_destination_violations_node = get_xml_node(xml_node, xml_tags.Elements.FLOW_SOURCE_VIOLATIONS,
                                                            optional=True)
            if flow_destination_violations_node is None:
                flow_destination_violations = None
            else:
                flow_destination_violations = FlowDestinationViolations.from_xml_node(flow_destination_violations_node)

            sources_in_zone = SourcesInZone.from_xml_node(xml_node)
            destinations_in_zone = DestinationsInZone.from_xml_node(xml_node)
            violating_services = ViolatingServices.from_xml_node(xml_node)

            return cls(severity, security_requirement, rule_properties_violations, flow_source_violations,
                       flow_destination_violations, sources_in_zone, destinations_in_zone, violating_services)


class ViolatingRule(XML_Object_Base):
    def __init__(self, rule, violations):
        super().__init__(xml_tags.Elements.VIOLATING_RULE)
        self.rule = rule
        self.violations = violations

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        rule = Rule.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.RULE))
        violations = XML_List.from_xml_node_by_tags(xml_node, xml_tags.Elements.VIOLATIONS, xml_tags.Elements.VIOLATION,
                                                    Violation)
        return cls(rule, violations)


class SecurityPolicyDeviceViolations(XML_Object_Base):
    def __init__(self, device_name, severity, violating_rules):
        super().__init__(xml_tags.Elements.SECURITY_POLICY_DEVICE_VIOLATIONS)
        self.device_name = device_name
        self.severity = severity
        self.violating_rules = violating_rules

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        device_name = get_xml_text_value(xml_node, xml_tags.Elements.DEVICE_NAME)
        severity = get_xml_text_value(xml_node, xml_tags.Elements.SEVERITY)
        violating_rules = XML_List.from_xml_node_by_tags(xml_node, xml_tags.Elements.VIOLATING_RULES,
                                                         xml_tags.Elements.VIOLATING_RULE, ViolatingRule)
        return cls(device_name, severity, violating_rules)


class TrafficRangeObject(XML_Object_Base):
    def __init__(self, tag, from_, to):
        """
        :type tag: str
        :type from_: str
        :type to: str
        """
        super().__init__(tag)
        self.from_ = from_
        self.to = to

    def __str__(self):
        if self.from_ == self.to:
            return str(self.from_)
        else:
            return "{}-{}".format(self.from_, self.to)

    def __repr__(self):
        return "{}({},{})".format(self.__class__.__name__, self.from_, self.to)

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        from_ = get_xml_text_value(xml_node, xml_tags.Elements.FROM)
        to = get_xml_text_value(xml_node, xml_tags.Elements.TO)
        return cls(from_, to)


class TrafficRangeSrc(TrafficRangeObject):
    def __init__(self, from_, to):
        super().__init__(xml_tags.Elements.SRC, from_, to)


class TrafficRangeDst(TrafficRangeObject):
    def __init__(self, from_, to):
        super().__init__(xml_tags.Elements.DST, from_, to)


class TrafficRangeProtocol(TrafficRangeObject):
    def __init__(self, from_, to):
        super().__init__(xml_tags.Elements.PROTOCOL, from_, to)


class TrafficRangePort(TrafficRangeObject):
    def __init__(self, from_, to):
        super().__init__(xml_tags.Elements.PORT, from_, to)


class TrafficRange(XML_Object_Base):
    def __init__(self, src, dst, protocol, port):
        """
        :type src: TrafficRangeSrc
        :type dst: TrafficRangeDst
        :type protocol: TrafficRangeProtocol
        :type port: TrafficRangePort
        """
        super().__init__(xml_tags.Elements.TRAFFIC_RANGE)
        self.src = src
        self.dst = dst
        self.protocol = protocol
        self.port = port

    def __repr__(self):
        return "TrafficRange('{}','{}','{}','{}')".format(self.src, self.dst, self.protocol, self.port)

    @property
    def tuple_header(self):
        return "Source", "Destination", "Protocol", "Port"

    def as_tuple(self):
        content = self.src, self.dst, self.protocol, self.port
        return self.tuple_header, content

    @classmethod
    def from_xml_node(cls, xml_node):
        """
        Initialize the object from a XML node.
        :param xml_node: The XML node from which all necessary parameters will be parsed.
        :type xml_node: xml.etree.Element
        """
        src = TrafficRangeSrc.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.SRC))
        dst = TrafficRangeDst.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.DST))
        protocol = TrafficRangeProtocol.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.PROTOCOL))
        port = TrafficRangePort.from_xml_node(get_xml_node(xml_node, xml_tags.Elements.PORT))
        return cls(src, dst, protocol, port)

    def as_netaddr_obj(self):
        if self.src.from_ == self.src.to:
            src = netaddr.IPNetwork(self.src.from_, flags=netaddr.ZEROFILL)
        else:
            src = netaddr.IPRange(self.src.from_, self.src.to, flags=netaddr.ZEROFILL)
        return src

    def as_netaddr_set(self):
        """This returns a netaddr set representing the TrafficRange"""
        return netaddr.IPSet(self.as_netaddr_obj())

    def as_service_type(self):
        if self.port.from_ == self.port.to:
            return Single_Service_Type(self.protocol.from_, self.port.from_)
        else:
            return Range_Service_Type(self.protocol.from_, self.port.from_, self.port.to)




import xml.sax.saxutils

from pytos.common.base_types import XML_Object_Base, XML_List
from pytos.common.definitions import xml_tags


class Step_Field_Base(XML_Object_Base):
    FIELD_CONTENT_ATTRIBUTES = None

    def __init__(self, num_id, name, read_only=None):
        super().__init__(xml_tags.Elements.FIELD)
        self.set_attrib(xml_tags.NAMESPACE_FIELD_ATTRIB_CONTENT, xml_tags.XSI_NAMESPACE_URL)
        self.id = num_id
        self.name = name
        self.read_only = read_only

    def get_field_type(self):
        try:
            return self._attribs[xml_tags.Attributes.XSI_TYPE]
        except KeyError:
            raise KeyError("Could not find type attribute, existing attributes: {}".format(self._attribs))

    def get_field_value(self):
        """
        Get the value(s) of a field.
        If the field type contains more than one content attribute, the output will be a dictionary.
        """
        attribute_names = self.__class__.FIELD_CONTENT_ATTRIBUTES
        if isinstance(attribute_names, list):
            values_dict = {}
            for attribute_name in attribute_names:
                attribute_value = self.__dict__.get(attribute_name)
                if isinstance(attribute_value, str):
                    values_dict[attribute_name] = xml.sax.saxutils.unescape(attribute_value)
                else:
                    values_dict[attribute_name] = attribute_value
            return values_dict
        else:
            if isinstance(self.__dict__[attribute_names], str):
                return xml.sax.saxutils.unescape(self.__dict__.get(attribute_names))
            else:
                return self.__dict__.get(attribute_names)

    def set_field_value(self, value):
        """
        Set the value(s) of a field.
        If the field type contains more than one content attribute, the input must be a dictionary.
        """
        attribute_names = self.__class__.FIELD_CONTENT_ATTRIBUTES
        if isinstance(attribute_names, list):
            if not isinstance(value, dict):
                raise TypeError("For field types with more than one content attribute, value must be a dict.")
            for attribute_name in attribute_names:
                try:
                    if isinstance(value[attribute_name], str):
                        self.__dict__[attribute_name] = xml.sax.saxutils.escape(value[attribute_name])
                    else:
                        self.__dict__[attribute_name] = value[attribute_name]
                except KeyError:
                    continue
        else:
            if isinstance(self.__dict__[attribute_names], str):
                self.__dict__[attribute_names] = xml.sax.saxutils.escape(value)
            else:
                self.__dict__[attribute_names] = value

    def from_xml_node(self, xml_node):
        raise NotImplementedError

    def __str__(self):
        field_value = self.get_field_value()
        field_string = "{}:\n".format(self.name)
        if field_value:
            if isinstance(field_value, dict):
                for key, value in field_value.items():
                    field_string += "\t{}: {}\n".format(key, value)
            else:
                field_string = "{}: {}\n".format(self.name, field_value)
        return field_string


class Step_Multi_Field_Base(Step_Field_Base):
    def __init__(self, num_id, name, read_only=None):
        super().__init__(num_id, name, read_only)

    def from_xml_node(self, xml_node):
        raise NotImplementedError

    def set_field_value(self, values):
        """
        Set the value(s) of a field.
        If the field type contains more than one content attribute, the input must be a dictionary.
        """
        attribute_names = self.__class__.FIELD_CONTENT_ATTRIBUTES
        if isinstance(attribute_names, list):
            if not isinstance(values, dict):
                raise TypeError("For field types with more than one content attribute, values must be a dict.")
            value_dict = {}
            for key, value in values.items():
                if isinstance(value, list):
                    escaped_values = []
                    for item in value:
                        if isinstance(item, str):
                            escaped_values.append(xml.sax.saxutils.escape(item))
                        else:
                            escaped_values.append(item)
                    value_dict[key] = escaped_values
                elif isinstance(value, str):
                    value_dict[key] = xml.sax.saxutils.escape(value)
                else:
                    value_dict[key] = value
            super().set_field_value(value_dict)
        else:
            if isinstance(values, XML_List):
                super().set_field_value(values)
            else:
                escaped_values = []
                for value in values:
                    if isinstance(value, str):
                        escaped_values.append(xml.sax.saxutils.escape(value))
                    else:
                        escaped_values.append(value)
                super().set_field_value(escaped_values)

    def get_field_value(self):
        attribute_names = self.__class__.FIELD_CONTENT_ATTRIBUTES
        if isinstance(attribute_names, list):
            value_dict = {}
            for key, value in super().get_field_value().items():
                if isinstance(value, str):
                    value_dict[key] = xml.sax.saxutils.escape(value)
                else:
                    value_dict[key] = value
            return value_dict
        else:
            unescaped_values = []
            field_value = super().get_field_value()
            try:
                if isinstance(field_value, XML_List):
                    return field_value
                else:
                    for item in super().get_field_value():
                        if isinstance(item, str):
                            unescaped_values.append(xml.sax.saxutils.unescape(item))
                        else:
                            unescaped_values.append(item)
            except TypeError:
                return field_value
            else:
                return unescaped_values

    def __str__(self):
        field_values = []
        field_string = "{}: \n".format(self.name)
        try:
            for value in self.get_field_value():
                try:
                    value = str(value).strip()
                except AttributeError:
                    pass
                field_values.append(value)
        except TypeError:
            pass
        field_values = [value for value in field_values if value]
        for index, value in enumerate(field_values):
            field_string += "\t{}. {}\n".format(index + 1, value)
        return field_string


class Target_Base(XML_Object_Base):
    def __init__(self, xml_tag, num_id, target_type=""):
        super().__init__(xml_tag)
        self.set_attrib(xml_tags.TYPE_ATTRIB, target_type)
        self.id = num_id

    def get_target_type(self):
        return self._attribs[xml_tags.TYPE_ATTRIB]

    def from_xml_node(self, xml_node):
        raise NotImplementedError


class Access_Request_Target(Target_Base):
    def __init__(self, xml_tag, num_id, target_type, region=None):
        self.region = region
        super().__init__(xml_tag, num_id, target_type)

    def to_pretty_str(self):
        raise NotImplementedError

    def from_xml_node(self, xml_node):
        raise NotImplementedError




import http.cookiejar
import logging
import multiprocessing

from pytos.common import rest_requests
from pytos.common.logging.definitions import HELPERS_LOGGER_NAME
from pytos.common.functions import get_local_ip_addresses

logger = logging.getLogger(HELPERS_LOGGER_NAME)


def verify_is_running_on_localhost(func):
    def func_wrapper(self, *args):
        if not self._is_running_on_localhost():
            raise IOError("This method can only be used when running on the local TOS installation.")
        else:
            return func(self, *args)

    return func_wrapper


class Secure_API_Helper:
    """
    This class is a base class that is used to interact via HTTP with the various TSS products.
    It also allows for easy sending of email messages.
    This class is not meant to be used on its own, but rather as a base class for the other helper classes, each
    specific to SecureTrack,SecureChange or SecureApp.

    :cvar LOGIN_COOKIE_TIMEOUT_SECONDS: The amount of time in seconds that the login cookie is valid for.
    """
    NUM_API_THREADS = multiprocessing.cpu_count()
    LOGIN_COOKIE_TIMEOUT_SECONDS = 120
    CONFIG_PARSER_SECTION_NAME = "securetrack"

    def __init__(self, hostname, login_data, timeout=None, max_retries=None):
        """Constructor

        :param hostname: The hostname with which we will communicate with HTTP.
        :type hostname: str
        :param login_data: A tuple of (username,password) used for basic authentication with the specified hostname.
        :type login_data: tuple
        :keyword timeout: (Optional) The timeout in seconds for HTTP requests.
        :type timeout: int
        :keyword max_retries: (Optional) The retries count of request failure.
        :type max_retries: int
        """
        self.hostname = hostname
        if self.hostname == "127.0.0.1":
            self._real_hostname = get_local_ip_addresses()[0]
        else:
            self._real_hostname = hostname
        self.login_data = {"username": login_data[0], "password": login_data[1]}
        self.cookie_jar = http.cookiejar.CookieJar()
        self._application_url = None
        self.max_retries = max_retries
        self.timeout = timeout

    def _is_running_on_localhost(self):
        if self.hostname.startswith("127.0.0.1") or self.hostname.startswith("localhost"):
            return True
        else:
            return False

    def get_uri(self, uri, expected_status_codes=None, headers=None, timeout=None, max_retries=None, session=None):
        """Make a GET request to a URI for the configured host.

        :param uri: The URI relative to the configured host to GET.
        :type uri: str
        :keyword expected_status_codes: (Optional) An integer or a list of integers representing HTTP status codes.
        :type expected_status_codes: int or list of ints
        :keyword timeout: (Optional) Set the timeout for the request.
        :type timeout: float
        :return: A response object for the GET request.
        """
        if headers is None:
            headers = {}
        expected_status_codes = expected_status_codes
        timeout = timeout
        if timeout is None:
            timeout = self.timeout
        if max_retries is None:
            max_retries = self.max_retries

        get_request = rest_requests.GET_Request(self._real_hostname, uri, login_data=self.login_data,
                                                headers=headers, verify_ssl=False,
                                                expected_status_codes=expected_status_codes,
                                                timeout=timeout, cookies=self.cookie_jar,
                                                session=session, max_retries=max_retries)
        return get_request

    def post_uri(self, uri, body=None, params_dict=None, multi_part_form_params=None, expected_status_codes=None,
                 cgi=None, headers=None, timeout=None, session=None):
        """Make a POST request to a URI for the configured host.

        :param headers: Additional headers to add to the request.
        :param uri: The URI relative to the configured host to POST.
        :type uri: str
        :param body: The string that will be used as the body for the POST request.
        :type body: str|bytes
        :param params_dict: A dictionary of params that will be used in the request body in the form of /URI?key=value.
        :type params_dict: dict
        :param multi_part_form_params: A tuple of parameters that will be encoded in multipart/form encoding.
            If the tuple contains 2 items, the first one will be used as the parameter name, the second will be the
            value.
            If the tuple contains 3 items, the first will be used as the parameter name, the second will be a open file
            handle, the third will be the name for the file to be sent.
        :type multi_part_form_params: tuple|int|dict
        :keyword expected_status_codes: (Optional) An integer or a list of integers representing HTTP status codes.
        :type expected_status_codes: int |list[int]
        :keyword cgi: (Optional) POST in CGI mode.
        :type cgi: bool
        :keyword timeout: (Optional) Set the timeout for the request.
        :type timeout: float
        :return: A response object for the POST request.
        """
        if headers is None:
            headers = {}
        timeout = timeout
        if timeout is None:
            timeout = self.timeout

        if params_dict:
            post_request = rest_requests.POST_Request(self._real_hostname, uri, body, cgi=cgi,
                                                      params=params_dict, headers=headers,
                                                      verify_ssl=False,
                                                      expected_status_codes=expected_status_codes,
                                                      timeout=timeout, cookies=self.cookie_jar,
                                                      session=session)
        else:
            if multi_part_form_params is not None:
                logger.debug("Sending multi part request, data is " + str(multi_part_form_params))
                post_request = rest_requests.POST_Request(self._real_hostname, uri, body, headers=headers,
                                                          multi_part_form_params=multi_part_form_params,
                                                          cgi=cgi, login_data=self.login_data,
                                                          verify_ssl=False,
                                                          expected_status_codes=expected_status_codes,
                                                          timeout=timeout, cookies=self.cookie_jar,
                                                          session=session)
            else:
                logger.debug("Sending regular POST.")
                post_request = rest_requests.POST_Request(self._real_hostname, uri, body, headers=headers,
                                                          login_data=self.login_data, verify_ssl=False,
                                                          expected_status_codes=expected_status_codes,
                                                          timeout=timeout, cookies=self.cookie_jar,
                                                          session=session, cgi=cgi)
        return post_request

    def put_uri(self, uri, body=None, expected_status_codes=None, headers=None, timeout=None,
                session=None):
        """Make a PUT request to a URI for the configured host.

        :param uri: The URI relative to the configured host to PUT.
        :type uri: str
        :param body: The string that will be used as the body for the PUT request.
        :type body: str|bytes
        :param params_dict: A dictionary of params that will be used in the request body in the form of /URI?key=value.
        :type params_dict: dict
        :keyword expected_status_codes: (Optional) An integer or a list of integers representing HTTP status codes.
        :type expected_status_codes: int| list[int]
        :keyword timeout: (Optional) Set the timeout for the request.
        :type timeout: float
        :return: A response object for the PUT request.
        """
        if headers is None:
            headers = {}
        timeout = timeout
        if timeout is None:
            timeout = self.timeout

        put_request = rest_requests.PUT_Request(self._real_hostname, uri, body, headers=headers,
                                                login_data=self.login_data, verify_ssl=False,
                                                expected_status_codes=expected_status_codes,
                                                timeout=timeout, cookies=self.cookie_jar,
                                                session=session)
        return put_request

    def delete_uri(self, uri, headers=None, session=None, **kwargs):
        """Make a DELETE request to a URI for the configured host.

        :param uri: The URI relative to the configured host to DELETE.
        :type uri: str
        :keyword expected_status_codes: (Optional) An integer or a list of integers representing HTTP status codes.
        :type expected_status_codes: int|list[int]
        :keyword timeout: (Optional) Set the timeout for the request.
        :type timeout: float
        :return: A response object for the DELETE request.
        """
        expected_status_codes = kwargs.get("expected_status_codes")
        if headers is None:
            headers = {}
        timeout = kwargs.get("timeout")
        if timeout is None:
            timeout = self.timeout

        delete_request = rest_requests.DELETE_Request(self._real_hostname, uri, headers=headers,
                                                      login_data=self.login_data, verify_ssl=False,
                                                      expected_status_codes=expected_status_codes,
                                                      timeout=timeout, cookies=self.cookie_jar,
                                                      session=session)
        return delete_request

    @classmethod
    def from_secure_config_parser(cls, secure_config_parser_obj, **kwlist):
        hostname = secure_config_parser_obj.get(cls.CONFIG_PARSER_SECTION_NAME, "hostname")
        username = secure_config_parser_obj.get_username(cls.CONFIG_PARSER_SECTION_NAME)
        password = secure_config_parser_obj.get_password(cls.CONFIG_PARSER_SECTION_NAME)
        return cls(hostname, (username, password), **kwlist)